cpdef Unitary_Grid(double complex[:,:,::1] H, double[::1] c_mins, double[::1] dcs, long[::1] c_bins)

cpdef Create_Interpolation_Cache(double complex[:,:,::1] U_grid, long[::1] grid_cum_prod, long[::1] c_bins)
cpdef Create_Sym_Interpolation_Cache(double complex[:,:,::1] U_grid2, long[::1] grid_cum_prod, long[::1] c_bins)



#cython: language_level=3
from unipolator.ui import *
from unipolator.sym_ui import *
from unipolator.hamiltonian_system import *
from unipolator.trotter_system import *
from unipolator.sym_trotter_system import *

# wrap
def test():
    print("test")


cpdef double creal(double complex dc) nogil

cpdef double cimag(double complex dc) nogil

cpdef double complex cconj(double complex dc) nogil

cpdef void cdiag(double complex[:,::1] A, double complex[::1] diagA) nogil

cpdef void Conj_mat_copy(double complex[:,::1] A, double complex[:,::1] ConjA) nogil

cpdef void Conj_mat(double complex[:,::1] A) nogil

cpdef void Conj_vec_copy(double complex[::1] A, double complex[::1] ConjA) nogil

cpdef void Conj_vec(double complex[::1] A) nogil

cpdef Dag(double complex[:,::1] A)
cpdef void Dag_fast(double complex[:,::1] A, double complex[:,::1] DagA) nogil
cdef double complex conj(double complex A) nogil
cpdef double abs_2(double complex A) nogil

cpdef void copy_v_exp_blas(double[::1] E, double complex[:,::1] V, double complex[:,::1] A, double t, int n) nogil
cpdef void v_exp_blas(double[::1] E, double complex[:,::1] V, double t) nogil

cdef void copy_pointer(double complex *v0, double complex *a0, int n2) nogil
cdef void copy_pointer_d(double *v0, double *a0, int n2) nogil
cdef void expE_V_expE_pointer(double *e0, double complex *v0, double t, double complex *expe0, int n) nogil
cdef void expE_V_expE(double[::1] E, double complex[:,::1] V, double t, double complex[::1] expE) nogil
cdef void d_dt_expE_V_expE_pointer(double amp, double *e0, double complex *v0, double complex *dv0, double t, double complex *expe0, int n) nogil
cdef void d_dt_expE_V_expE(double amp, double[::1] E, double complex[:,::1] V, double complex[:,::1] dV, double t, double complex[::1] expE) nogil
cdef void v_exp_pointer(double *e0, double complex *v0, double t, int n) nogil
cdef void v_exp_v_pointer(double amp, double *e0, double complex *v0, double t, int n) nogil
cdef void d_dt_exp_v_pointer(double amp, double *e0, double complex *v0, double complex *dv0, double t, int n, int n_2) nogil

cdef void phase_shift_matrix_pointer(double *e0, double complex *v0, double complex *a0, double dt, int n) nogil
cpdef void phase_shift_matrix(double[::1] E, double complex[::1] v, double complex[:,::1] A, double dt) nogil

cdef void V_expE_W_pointer(double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, int n, int n2) nogil
cpdef void V_expE_W(double complex[:,::1] V, double complex[:,::1] W, double[::1] E, double t, double complex[:,::1] A, double complex[:,::1] B, int n, int n_2)
cdef void V_EexpE_W_pointer(double amp, double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, int n, int n2) nogil
cdef void d_dt_V_expE_W_pointer(double amp, double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, double complex *db0, int n, int n_2) nogil
cdef void DagV_expE_V_pointer(double complex *v0, double *e0, double t, double complex *a0, double complex *b0, int n, int n_2) nogil
cpdef void c_expmH(double complex[:,::1] H, double dt, double complex[:,::1] U, int lwork, int lrwork, int liwork)

cpdef void eigU(double complex[:,::1] U, double complex[::1] Ec, double complex[:,::1] V)

cpdef double complex clog(double complex c) nogil

cpdef void copy_v_log_blas(double complex[::1] E, double complex[:,::1] V, double complex[:,::1] A)

cpdef void logmU(double complex[:,::1] U, double complex[:,::1] logU, double complex[:,::1] Emat, double complex[::1] E, double complex[:,::1] V)
cpdef void ilogmU_eig(double complex[:,::1] U, double complex[::1] E, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V)

cpdef void Make_U_Partial(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V)
cpdef void Partialize_Center(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat2, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V)
cpdef void Partialize_Sides(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V_L, double complex[:,::1] V_R)



#cython: language_level=3
from cython.operator import dereference
from libc.math cimport cos, sin, log, atan2, sqrt, fabs
import numpy as np
#cimport numpy as np
from scipy.linalg import schur
from scipy.linalg.cython_blas cimport zgemm, zscal, zdscal, zaxpy, zcopy, dcopy
from scipy.linalg.cython_lapack cimport zheevd
from .blas_functions cimport MM_cdot, DagM_M_cdot, M_DagM_cdot, MM_cdot_pointer, DagM_M_cdot_pointer, c_eigh_lapack

##### Basic complex number functions to avoid using complex.h #########################
cpdef double creal(double complex dc) nogil:
    cdef double complex* dcptr = &dc
    return (<double *>dcptr)[0]

cpdef double cimag(double complex dc) nogil:
    cdef double complex* dcptr = &dc
    return (<double *>dcptr)[1]

cpdef double complex cconj(double complex dc) nogil:
    dc.imag = -dc.imag 
    #a =  (<double *>dcptr)[0] -1j*(<double *>dcptr)[1]
    return dc

cpdef void cdiag(double complex[:,::1] A, double complex[::1] diagA) nogil:
    cdef int n = A.shape[0]
    cdef int i
    for i in range(n):
        diagA[i] = A[i,i]

cpdef void Conj_mat_copy(double complex[:,::1] A, double complex[:,::1] ConjA) nogil:
    cdef int n = A.shape[0]
    cdef int i,j
    for i in range(n):
        for j in range(n):
            ConjA[i,j].real =  A[i,j].real
            ConjA[i,j].imag = -A[i,j].imag

cpdef void Conj_mat(double complex[:,::1] A) nogil:
    cdef int n = A.shape[0]
    cdef int i,j
    for i in range(n):
        for j in range(n):
            A[i,j].imag = -A[i,j].imag

cpdef void Conj_vec_copy(double complex[::1] A, double complex[::1] ConjA) nogil:
    cdef int n = A.shape[0]
    cdef int i
    for i in range(n):
        ConjA[i].real =  A[i].real
        ConjA[i].imag = -A[i].imag

cpdef void Conj_vec(double complex[::1] A) nogil:
    cdef int n = A.shape[0]
    cdef int i
    for i in range(n):
        A[i].imag = -A[i].imag


cpdef Dag(double complex[:,::1] A):
    cdef int n = A.shape[0]
    cdef int i, j
    cdef double complex[:,::1] DagA = np.empty([n,n], dtype=np.complex128)
    for i in range(n):
        for j in range(n):
            DagA[i,j].real =  A[j,i].real
            DagA[i,j].imag = -A[j,i].imag
    return DagA

cpdef void Dag_fast(double complex[:,::1] A, double complex[:,::1] DagA) nogil:
    cdef int n = A.shape[0]
    cdef int i, j
    for i in range(n):
        for j in range(n):
            DagA[i,j].real =  A[j,i].real
            DagA[i,j].imag = -A[j,i].imag

cdef double complex conj(double complex A) nogil:
    cdef double complex conjA
    conjA.real = A.real
    conjA.imag = -A.imag
    return conjA

cpdef double abs_2(double complex A) nogil: #c_dagc
    cdef double absA_2
    absA_2 = A.real * A.real + A.imag * A.imag
    return absA_2

##### Add Commutator function !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#### Vector exponential/logarithm times matrix

# Old skripts for this
cpdef void copy_v_exp_blas(double[::1] E, double complex[:,::1] V, double complex[:,::1] A, double t, int n) nogil:
    #cdef int n = V.shape[0]
    cdef double complex c
    cdef double ei
    cdef double complex *vi0
    cdef int incz = 1
    cdef int i
    A[:,:] = V
    for i in range(n):
        ei = E[i]*t
        c = cos(ei)-1j*sin(ei) #cexp(ei)
        vi0=&A[i,0]
        zscal(&n, &c, vi0, &incz)
cpdef void v_exp_blas(double[::1] E, double complex[:,::1] V, double t) nogil:
    cdef int n = V.shape[0]
    cdef double complex c
    cdef double ei
    cdef double complex *vi0
    cdef int incz = 1
    cdef int i
    for i in range(n):
        ei = E[i]*t
        c = cos(ei)-1j*sin(ei) #cexp(ei)
        vi0=&V[i,0]
        zscal(&n, &c, vi0, &incz)

cdef void copy_pointer(double complex *v0, double complex *a0, int n2) nogil:
    #cdef int n2 = n*n
    cdef int incz = 1
    zcopy(&n2, v0, &incz, a0, &incz)


cdef void copy_pointer_d(double *v0, double *a0, int n2) nogil:
    #cdef int n2 = n*n
    cdef int incz = 1
    dcopy(&n2, v0, &incz, a0, &incz)

# New skripts for this
cdef void expE_V_expE_pointer(double *e0, double complex *v0, double t, double complex *expe0, int n) nogil:
    # Construct the transformation:  A <- diag(exp(v)) @ A @ diag(exp(v))
    ## This should be auto simd
    cdef double complex c, d
    cdef double ei
    cdef int incz = 1
    cdef int i, j
    cdef double complex *expe0_0 = expe0 # .copy()
    for i in range(n):
        ei = dereference(e0) * t
        c = cos(ei) -1j*sin(ei) #cexp(ei)
        expe0[0] = c
        e0 += 1
        expe0 += 1
    for i in range(n):
        c = expe0_0[0]
        expe0_0 += 1
        expe0 -= n
        for j in range(n):
            d = c * expe0[0]
            v0[0] = d * v0[0]
            expe0 += 1
            v0 += 1
cdef void expE_V_expE(double[::1] E, double complex[:,::1] V, double t, double complex[::1] expE) nogil:
    cdef int n = V.shape[0]
    cdef double *e0 = &E[0]
    cdef double complex *expe0 = &expE[0]
    cdef double complex *v0 = &V[0,0]
    expE_V_expE_pointer(e0, v0, t, expe0, n)

cdef void d_dt_expE_V_expE_pointer(double amp, double *e0, double complex *v0, double complex *dv0, double t, double complex *expe0, int n) nogil:
    # Construct the derivative of the transformation:  A <- diag(exp(v)) @ A @ diag(exp(v))
    ## This should be auto simd
    cdef double complex c, d, ec, s
    cdef double ei
    cdef int incz = 1
    cdef int i, j
    cdef double complex *expe0_0 = expe0 # .copy()
    cdef double *e0_0 = e0 # .copy()
    cdef double complex iamp = -1j * amp
    for i in range(n):
        ei = dereference(e0) * t
        c.real = cos(ei)
        c.imag = -sin(ei) #cexp(ei)
        expe0[0] = c
        e0 += 1
        expe0 += 1
    for i in range(n):
        c = expe0_0[0]
        expe0_0 += 1
        ec = e0_0[0]
        e0_0 += 1
        expe0 -= n
        e0 -= n
        for j in range(n):
            d = c * expe0[0]
            s = iamp * (ec + e0[0])
            v0[0] = d * v0[0]
            dv0[0] = s * v0[0]
            expe0 += 1
            e0 += 1
            v0 += 1
            dv0 += 1
cdef void d_dt_expE_V_expE(double amp, double[::1] E, double complex[:,::1] V, double complex[:,::1] dV, double t, double complex[::1] expE) nogil:
    cdef int n = V.shape[0]
    cdef double *e0 = &E[0]
    cdef double complex *expe0 = &expE[0]
    cdef double complex *v0 = &V[0,0]
    cdef double complex *dv0 = &dV[0,0]
    d_dt_expE_V_expE_pointer(amp, e0, v0, dv0, t, expe0, n)

cdef void v_exp_pointer(double *e0, double complex *v0, double t, int n) nogil:
    cdef double complex c
    cdef double ei
    cdef int incz = 1
    cdef int i

    ei = dereference(e0) * t
    c.real = cos(ei)
    c.imag = -sin(ei)
    zscal(&n, &c, v0, &incz)
    for i in range(1, n):
        v0 += n
        e0 += 1
        ei = dereference(e0) * t
        c.real = cos(ei)
        c.imag = -sin(ei)
        zscal(&n, &c, v0, &incz)
cdef void v_exp_v_pointer(double amp, double *e0, double complex *v0, double t, int n) nogil: # -i*E*amp*exp(-i*E*t)
    cdef double complex c
    cdef double ei, f
    cdef int incz = 1
    cdef int i

    ei = dereference(e0) * t
    f = -dereference(e0) * amp
    c.real = f * sin(ei)
    c.imag = f * cos(ei)
    zscal(&n, &c, v0, &incz)
    for i in range(1, n):
        v0 += n
        e0 += 1
        ei = dereference(e0) * t
        f = -dereference(e0) * amp
        c.real = f * sin(ei)
        c.imag = f * cos(ei)
        zscal(&n, &c, v0, &incz)
cdef void d_dt_exp_v_pointer(double amp, double *e0, double complex *v0, double complex *dv0, double t, int n, int n_2) nogil: # -i*E*amp*exp(-i*E*t)
    cdef double complex c, dc
    cdef double ei, f
    cdef int incz = 1
    cdef int i
    copy_pointer(v0, dv0, n_2)
    ei = dereference(e0) * t
    f = -dereference(e0) * amp
    c.real = cos(ei)
    c.imag = -sin(ei)
    dc.real = f * c.real
    dc.imag = - f * c.imag
    zscal(&n, &c, v0, &incz)
    zscal(&n, &dc, dv0, &incz)
    for i in range(1, n):
        v0 += n
        dv0 += n
        e0 += 1
        ei = dereference(e0) * t
        f = -dereference(e0) * amp
        c.real = cos(ei)
        c.imag = -sin(ei)
        dc.real = f * c.real
        dc.imag = - f * c.imag
        zscal(&n, &c, v0, &incz)
        zscal(&n, &dc, dv0, &incz)

cdef void phase_shift_matrix_pointer(double *e0, double complex *v0, double complex *a0, double dt, int n) nogil: # Only calculates upper triangle of symmetric matrix
    cdef int i, j
    cdef double complex idt = - 1j*dt
    cdef double ei, de
    cdef double complex *a1 = a0
    ei = dereference(e0) * dt
    v0[0].real = cos(ei)
    v0[0].imag = -sin(ei)
    for i in range(1, n):
        e0 += 1
        v0 += 1
        ei = dereference(e0) * dt
        v0[0].real = cos(ei)
        v0[0].imag = -sin(ei)
    e0 -= n - 1
    v0 -= n - 1
    for i in range(n):
        # Diagonal case
        #print(idt * v0[i])
        a0[0] = idt * v0[i]
        for j in range(i+1, n):
            a0 += 1
            a1 += n
            de = e0[i] - e0[j]
            if fabs(de)< 10**-14:
                a0[0] = idt * v0[i]
            else:
                a0[0] = 1/de * (v0[i] - v0[j])
            a1[0] = a0[0]
        a0 += i+2
        a1 = a0
cpdef void phase_shift_matrix(double[::1] E, double complex[::1] v, double complex[:,::1] A, double dt) nogil: # Only calculates upper triangle of symmetric matrix
    cdef double *e0 = &E[0]
    cdef double complex *v0 = &v[0]
    cdef double complex *a0 = &A[0,0]
    cdef int n = A.shape[0]
    phase_shift_matrix_pointer(e0, v0, a0, dt, n)



##### Add differential!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

###### Matrix exponentials from eigenvalue decomposition of Hamiltonians and Unitary logarithms / decompositions

cdef void V_expE_W_pointer(double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, int n, int n_2) nogil:
    # Calculates B = V @ diag(exp(-1j*E*t)) @ W
    copy_pointer(w0 , a0, n_2)
    v_exp_pointer(e0, a0, t, n)
    MM_cdot_pointer(v0, a0, b0, n)
cpdef void V_expE_W(double complex[:,::1] V, double complex[:,::1] W, double[::1] E, double t, double complex[:,::1] A, double complex[:,::1] B, int n, int n_2):
    # Calculates B = V @ diag(exp(-1j*E*t)) @ W
    cdef double *e0 = &E[0]
    cdef double complex *v0 = &V[0,0]
    cdef double complex *w0 = &W[0,0]
    cdef double complex *a0 = &A[0,0]
    cdef double complex *b0 = &B[0,0]
    V_expE_W_pointer(v0, w0, e0, t, a0, b0, n, n_2)

cdef void V_EexpE_W_pointer(double amp, double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, int n, int n_2) nogil:
    # Calculates amp*dB/dt = -1j * amp @ V @ diag(E) * diag(exp(-1j*E*t)) @ W   -> with helper variable *a0
    copy_pointer(w0 , a0, n_2)
    v_exp_v_pointer(amp, e0, a0, t, n)
    MM_cdot_pointer(v0, a0, b0, n)
cdef void d_dt_V_expE_W_pointer(double amp, double complex *v0, double complex *w0, double *e0, double t, double complex *a0, double complex *b0, double complex *db0, int n, int n_2) nogil:
    # Calculates B = V @ diag(exp(-1j*E*t)) @ W    -> with helper variable *a0
    # and its derivative amp*dB/dt = -1j * amp @ V @ diag(E) * diag(exp(-1j*E*t)) @ W
    copy_pointer(w0, a0, n_2)
    d_dt_exp_v_pointer(amp, e0, a0, b0, t, n, n_2)
    MM_cdot_pointer(v0, b0, db0, n)
    MM_cdot_pointer(v0, a0,  b0, n)
cdef void DagV_expE_V_pointer(double complex *v0, double *e0, double t, double complex *a0, double complex *b0, int n, int n_2) nogil:
    # Calculates B = Dag(V) @ diag(exp(-1j*E*t)) @ V
    copy_pointer(v0 , a0, n_2)
    v_exp_pointer(e0, a0, t, n)
    DagM_M_cdot_pointer(v0, a0, b0, n)

cpdef void c_expmH(double complex[:,::1] H, double dt, double complex[:,::1] U, int lwork, int lrwork, int liwork):
    cdef int n = H.shape[0]
    cdef double complex *h0=&H[0,0]
    cdef double complex[:,::1] V = np.empty([n, n], dtype=np.complex128)
    cdef double complex[:,::1] R = np.empty([n, n], dtype=np.complex128)
    cdef double complex *v0=&V[0,0]
    cdef double complex *r0=&V[0,0]
    cdef double[::1] E = np.empty([n], dtype=np.double)
    cdef double *e0=&E[0]
    c_eigh_lapack(H, V, E , lwork, lrwork, liwork)
    copy_v_exp_blas(E, V, R, dt, n)
    DagM_M_cdot(V, R, U)

cpdef void eigU(double complex[:,::1] U, double complex[::1] Ec, double complex[:,::1] V): # alternative to np.eig returning unitary matrices V
    cdef int n = U.shape[0]
    cdef int i, j
    Emat, Vmat = schur(U, output='complex')
    for i in range(n):
        Ec[i] = Emat[i, i]
        for j in range(n):
            V[i,j].real =  Vmat[j,i].real
            V[i,j].imag = -Vmat[j,i].imag

cpdef double complex clog(double complex c) nogil:   # logarithm of complex number
    cdef double x = c.real
    cdef double y = c.imag
    cdef double complex d
    cdef double squares = x*x + y*y
    cdef double abs_val = sqrt(squares)
    d.real = log(abs_val)
    d.imag = atan2(y,x)
    return d

cpdef void copy_v_log_blas(double complex[::1] E, double complex[:,::1] V, double complex[:,::1] A):
    cdef int n = E.shape[0]
    cdef double complex c
    cdef int i
    cdef double complex *vi0
    cdef int incz = 1
    A[:,:] = V
    for i in range(n):
        c = 1j*clog(E[i])
        vi0=&A[i,0]
        zscal(&n, &c, vi0, &incz)

cpdef void logmU(double complex[:,::1] U, double complex[:,::1] logU, double complex[:,::1] Emat, double complex[::1] E, double complex[:,::1] V):# Helper variables Emat, E, V
    cdef double dt = 1.0
    eigU(U, E, V)
    copy_v_log_blas(E, V, Emat)
    DagM_M_cdot(V, Emat, logU)

cpdef void ilogmU_eig(double complex[:,::1] U, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V):# Helper variables E, output variables: logE, V
    cdef int s = Ec.shape[0]
    cdef int i, j
    eigU(U, Ec, V_tmp)
    for i in range( s ):
        logE_tmp[i] = -clog(Ec[i]).imag
    # Sort the values
    cdef long long[::1] order
    order = np.argsort(logE_tmp)
    for i in range( s ):
        j = order[i]
        logE[i] = logE_tmp[j]
        V[i,:] = V_tmp[j,:]


cpdef void Make_U_Partial(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V):
    # Input matrices:                   U_1 @ Dag(U_0) -> using full steps U_0, U_1   ### double complex[:,::1] logU,
    # Intermediate matrices (vectors):  lugU, Emat, (Ec)
    # Result matrices (vectors):        V, (E)
    cdef int s = U_1.shape[0]
    M_DagM_cdot(U_1, U_0, Emat)
    ilogmU_eig(Emat, Ec, logE_tmp, V_tmp, logE, V)

cpdef void Partialize_Center(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat2, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V):
    # Input matrices:                   Dag(U_0) @ U_1^2 @ Dag(U_0) -> using half steps U_0, U_1
    # Intermediate matrices (vectors):  lugU, Emat, (Ec)
    # Result matrices (vectors):        V, (E)
    MM_cdot(U_1, U_1, V)  # lugU = U_1 ** 2
    DagM_M_cdot(U_0, V, Emat)
    M_DagM_cdot(Emat, U_0, Emat2)
    ilogmU_eig(Emat2, Ec, logE_tmp, V_tmp, logE, V)

cpdef void Partialize_Sides(double complex[:,::1] U_1, double complex[:,::1] U_0, double complex[:,::1] Emat, double complex[::1] Ec, double[::1] logE_tmp, double complex[:,::1] V_tmp, double[::1] logE, double complex[:,::1] V_L, double complex[:,::1] V_R):
    # Input matrices:                   Dag(U_0) @ U_1     &     U_1 @ Dag(U_0)    ->    using half steps U_0, U_1
    # Intermediate matrices (vectors):  lugU, Emat, V_tmp, (Ec)
    # Result matrices (vectors):        V_L, V_R, (E_L, E_R)
    DagM_M_cdot(U_0, U_1, Emat)
    ilogmU_eig(Emat, Ec, logE, V_tmp, logE, V_L)
    M_DagM_cdot(U_1, U_0, Emat)
    ilogmU_eig(Emat, Ec, logE_tmp, V_tmp, logE, V_R)


cdef void AxB_elementwise_pointer(double complex *a0, double complex *b0, double complex *c0, int n2) nogil

cdef void MM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil
cdef void MM_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil
cpdef void MM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil

cdef void MMM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, double complex *e0, int n) nogil
cdef void MMM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D, double complex[:,::1] E) nogil

cdef void DagM_M_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil
cdef void DagM_M_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil
cpdef void DagM_M_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil

cpdef void M_DagM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil
cdef void M_DagM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil
cdef void M_DagM_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil

# Triple products
cpdef void DagA_B_A_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D) nogil
cdef void DagA_B_A_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, int n) nogil
cpdef void A_B_DagA_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D) nogil
cdef void A_B_DagA_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, int n) nogil

cdef double complex tr_dot_pointer(double complex *v0, double complex *w0, int n) nogil
cpdef double complex tr_dot(double complex[:,::1] V, double complex[:,::1] W, int n )
cdef double complex tr_dot_pointer_target_indexes(double complex *v0, double complex *w0, int n, int[::1] target_indexes) nogil
cpdef double complex tr_dot_target_indexes(double complex[:,::1] V, double complex[:,::1] W, int n, int[::1] target_indexes )
cdef double complex target_indexes_trace_pointer(double complex *v0, int n, int[::1] target_indexes) nogil

cpdef (int, int, int) c_eigh_lapack_workspace_sizes(double complex[:,::1] H)

cpdef void c_eigh_lapack(double complex[:,::1] H, double complex[:,::1] V, double[::1] E, int lwork, int lrwork, int liwork)

cpdef void d_third_order_tensor_scale(double complex[:,:,::1] A, double d) nogil

cpdef void d_mat_scale(double complex[:,::1] A, double d) nogil

cpdef void c_mat_scale(double complex[:,::1] A, double complex c) nogil

cpdef void d_mat_add_first(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double d) # nogil:

cpdef void c_mat_add_first(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex c) # nogil:

cpdef void d_mat_add(double complex[:,::1] A, double complex[:,::1] B, double d) nogil
cdef void d_mat_add_pointer(double complex *a0, double complex *b0, double d, int n2) nogil

cpdef void c_mat_add(double complex[:,::1] A, double complex[:,::1] B, double complex c) nogil



#cython: language_level=3
from libc.math cimport fabs
import numpy as np
cimport numpy as npc

#### Construct Parameters of bounds #############################################################

cpdef Bin_Parameters(double[::1] c_mins, double[::1] c_maxs, long[::1] c_bins):
    # Correct input parameters
    cdef int n = c_mins.shape[0]
    cdef int i
    cdef double c_min, c_max, dc, ac
    cdef int c_bin
    cdef npc.ndarray[npc.double_t, ndim=1] dcs = np.empty(n, dtype=np.double)
    cdef npc.ndarray[npc.double_t, ndim=1] das = np.empty(n, dtype=np.double)
    cdef npc.ndarray[npc.double_t, ndim=1] c_minis = np.empty(n, dtype=np.double)
    cdef npc.ndarray[npc.double_t, ndim=1] c_maxis = np.empty(n, dtype=np.double)
    for i in range(n):
        if c_mins[i] <= c_maxs[i]:
            c_min = c_mins[i]
            c_max = c_maxs[i]
        else:
            c_min = c_maxs[i]
            c_max = c_mins[i]
        c_bin = c_bins[i]
        if abs(c_max -c_min) < 10**-15:
            print('This parameter might not be necessary! (c_max - c_min < 10**-15)')
            c_max = c_max + 10**-15
        dc = (c_max-c_min) / c_bin
        ac = c_bin / (c_max-c_min)
        c_minis[i] = c_min
        c_maxis[i] = c_max
        dcs[i] = dc
        das[i] = ac # For scaling the derivatives
    return c_minis, c_maxis, dcs, das

##### Iterate over multiple dimensions
cpdef void next_in_mrange(long[:] A, long[:] n_s) nogil:
    cdef int i = n_s.shape[0] - 1
    while i >= 0:
        A[i] = (A[i]+1) % n_s[i]
        if A[i]:
            break
        i += -1

##### Index Transformation to (partially) flattened arrays
cpdef reverse_cum_prod(long[:] s, int n):
    if n < 0:
        n = s.shape[0]
    cdef long[:] s_red = s[:n]
    cdef npc.ndarray[long, ndim=1] cum_prod = np.empty(n, dtype=long)
    cdef int curr_index = n-1
    cum_prod[curr_index] = 1 #s_red[-1]
    for i in range(n-1):
        curr_index += -1
        cum_prod[curr_index] = cum_prod[curr_index+1]*s_red[curr_index+1]
    cdef int tot_prod = cum_prod[0]*s[0]
    return cum_prod, tot_prod

cpdef int findex_0(long[:] i, long[:] cum_prod, int n): # n is length of i, cum_prod #### Inline this function and similar functions
    cdef int i_s
    cdef int a = i[0] * cum_prod[0]
    for i_s in range(1,n):
        a += i[i_s]*cum_prod[i_s]
    return a

cpdef int findex(long[:] i, long[:] cum_prod, int n, int a) nogil: # n is length of i, cum_prod #### Inline this function and similar functions
    # in this variant a is initialized via an a_min
    cdef int i_s
    for i_s in range(n):
        a += i[i_s]*cum_prod[i_s]
    return a

# Convert indices to flattened arrays (works equivalently to ndarrays)
cpdef flat_index_sizes_E(long[::1] c_bins):   # For eigene-enery vectors
    # Construct indexes for a flattened cache
    cdef int l = c_bins.shape[0]
    cdef int i, j
    cdef int tot_prod0 = 1
    cdef int tot_ind
    cdef long[::1] tot_prods = np.empty(l, dtype = long)
    cdef long[::1] d_bins = np.empty(l, dtype = long)
    cdef long[::1] first_elements = np.empty(l, dtype = long)
    cdef long[:,::1] array_sizes = np.empty([l,l], dtype = long)
    cdef long[:,::1] array_strides = np.empty([l,l], dtype = long)
    for i in range(l):
        d_bins[i] = c_bins[i] + 1 # Should copy data
        tot_prod0 = tot_prod0*d_bins[i]
    # Construct the total sizes of the caches for every dimensions
    for i in range(l): # One arrow cache for every direction in grid
        d_bins[i] = c_bins[i] # Should copy data
        if i > 0:
            d_bins[i-1] = c_bins[i-1]+1
        tot_prods[i] =  (tot_prod0*c_bins[i]) // (c_bins[i]+1)
        curr_cum_prod, _ = reverse_cum_prod(d_bins, l)
        for j in range(l):
            array_strides[i,j] = curr_cum_prod[j]
        for j in range(l):
            array_sizes[i,j] = d_bins[j]
    first_elements[0] = 0
    for i in range(1,l):
        first_elements[i] = first_elements[i-1]+tot_prods[i-1]
    tot_ind = first_elements[l-1]+tot_prods[l-1]
    return tot_ind, first_elements, array_strides, array_sizes

cpdef flat_index_sizes_V(long[::1] c_bins): # For orthonormal matrices and their products (left to center (low-high) to right)
    # Construct indexes for a flattened cache
    cdef int l = c_bins.shape[0]
    cdef int i, j
    cdef int tot_prod0 = 1
    cdef long[::1] d_bins = np.empty(l, dtype = long)
    cdef int tot_ind_L, tot_int_R, tot_ind_C

    cdef long[::1] tot_prods_C = np.empty(l-1, dtype = long)
    cdef long[::1] first_elements_C = np.empty(l-1, dtype = long)
    cdef long[:,::1] array_sizes_C = np.empty([l-1,l], dtype = long)
    cdef long[:,::1] array_strides_C = np.empty([l-1,l], dtype = long)

    cdef long[::1] array_sizes_L = np.empty(l, dtype = long)
    cdef long[::1] array_strides_L = np.empty(l, dtype = long)
    cdef long[::1] array_sizes_R = np.empty(l, dtype = long)
    cdef long[::1] array_strides_R = np.empty(l, dtype = long)
    for i in range(l):
        d_bins[i] = c_bins[i] + 1 # Should copy data
        tot_prod0 = tot_prod0*d_bins[i]
    # Construct the total sizes of the caches for every dimensions
    for i in range(l): # One arrow cache for every direction in grid
        tot_prod0 = (tot_prod0 * c_bins[i]) // (c_bins[i]+1)
        d_bins[i] = c_bins[i] # Should copy data
        if i > 1:
            d_bins[i-2] = c_bins[i-2]+1
            tot_prod0 = (tot_prod0 * (c_bins[i-2]+1)) // (c_bins[i-2])
        if i == 0: # R
            tot_ind_R = tot_prod0
            curr_cum_prod, _ = reverse_cum_prod(d_bins, l)
            for j in range(l):
                array_strides_R[j] = curr_cum_prod[j]
                array_sizes_R[j] = d_bins[j]
        elif i > 0: # C
            tot_prods_C[i-1] = tot_prod0
            curr_cum_prod, _ = reverse_cum_prod(d_bins, l)
            for j in range(l):
                array_strides_C[i-1, j] = curr_cum_prod[j]
                array_sizes_C[i-1, j] = d_bins[j]
        if i == l - 1: # L
            if i > 0:
                tot_prod0 = ( tot_prod0 * (c_bins[i-1]+1) ) // (c_bins[i-1])
                d_bins[i-1] = c_bins[i-1]+1
            tot_ind_L = tot_prod0
            curr_cum_prod, _ = reverse_cum_prod(d_bins, l)
            for j in range(l):
                array_strides_L[j] = curr_cum_prod[j]
                array_sizes_L[j] = d_bins[j]
    if l > 1:
        first_elements_C[0] = 0
        for i in range(1,l-1):
            first_elements_C[i] = first_elements_C[i-1] + tot_prods_C[i-1]
        tot_ind_C = first_elements_C[l-2]+tot_prods_C[l-2]
    else:
        tot_ind_C = 0
    return tot_ind_L, array_strides_L, array_sizes_L, tot_ind_C, first_elements_C, array_strides_C, array_sizes_C, tot_ind_R, array_strides_R, array_sizes_R

##### Interpolation Parameters ########################################################################
cpdef int is_odd(int num) nogil:
    return (num & 0x1) #bitwise and with hexadecimal 1

cpdef void is_odd_array(long[::1] num, int n) nogil:
    #cdef int n = num.shape[0]
    cdef int i
    for i in range(n):
        num[i] = (num[i] & 0x1)

cpdef int is_even(int num) nogil:
    return 1-(num & 0x1)

cpdef void is_even_array(long[::1] num, int n) nogil:
    #cdef int n = num.shape[0]
    cdef int i
    for i in range(n):
        num[i] = 1-(num[i] & 0x1)

cpdef void asym_sign(double[::1] nums, long[::1] signs, int n) nogil:
    #cdef int n = nums.shape[0]
    cdef int i
    for i in range(n):
        signs[i] = 1 if nums[i] >= 0 else -1

cpdef void asym_sign_2D(double[:,::1] nums, long[:,::1] signs, int n0, int n1) nogil:
    #cdef int n0 = nums.shape[0]
    #cdef int n1 = nums.shape[1]
    cdef int i, j
    for i in range(n0):
        for j in range(n1):
            signs[i,j] = 1 if nums[i,j] >= 0 else -1

cpdef void rounder(double a, int b) nogil:
    b = <int> a
    if a - b > 0.5:
        b += 1

cpdef void rounder_array(double[::1] a, long[::1] b, int n) nogil:
    #cdef int n = a.shape[0]
    cdef int i
    for i in range(n):
        b[i] = <int> a[i]
        if a[i] - b[i] > 0.5:
            b[i] += 1

cpdef void rounder_array_2d(double[:,::1] a, long[:,::1] b, int n0, int n1) nogil:
    #cdef int n0 = a.shape[0]
    #cdef int n1 = a.shape[1]
    cdef int i, j
    for i in range(n0):
        for j in range(n1):
            b[i,j] = <int> a[i,j]
            if a[i,j] - b[i,j] > 0.5:
                b[i,j] += 1

cpdef int is_in_interval(double[::1] c, long[::1] maxs, int n) nogil:
    cdef int inside = 1
    #cdef int n = c.shape[0]
    for i in range(n):
        if c[i] < 0 or c[i] > maxs[i]:
            inside = 0
            break
    return inside

cpdef int int_sum(long[::1] i_s, int n) nogil:
    #cdef int n = i_s.shape[0]
    cdef int s_i = 0
    cdef int i
    for i in range(n):
        s_i += i_s[i]
    return s_i

cpdef int int_max(long[::1] i_s, int n) nogil:
    #cdef int n = i_s.shape[0]
    cdef int m_i
    cdef int i
    if n > 0:
        m_i= i_s[0]
        for i in range(1,n):
            if i_s[i] > m_i:
                m_i = i_s[i]
    else:
        m_i = 0
    return m_i

cpdef int int_prod(long[::1] i_s, int n) nogil:
    #cdef int n = i_s.shape[0]
    cdef int p_i = i_s[0]
    cdef int i
    for i in range(1,n):
        p_i = p_i*i_s[i]
    return p_i

cpdef int_prod_array(long[:,::1] i_s, int n): # square
    cdef int i, j
    cdef long[::1] p_i = np.empty(i_s.shape[0], dtype=long)
    for i in range(n):
        p_i[i] = i_s[i, 0]
        for j in range(1,n):
            p_i[i] *= i_s[i, j]
    return p_i

cpdef int_prod_asym_array(long[:,::1] i_s): # square
    cdef int i, j
    cdef int[::1] p_i = np.empty(i_s.shape[0], dtype=int)
    for i in range(i_s.shape[0]):
        p_i[i] = i_s[i, 0]
        for j in range(1,i_s.shape[1]):
            p_i[i] *= i_s[i, j]
    return p_i

cpdef void c_abs_sum(double[::1] c, double[::1] abs_c, int n) nogil:
    cdef int i
    for i in range(n):
        abs_c[i] = fabs(c[i])

cpdef void c_int_sum(long[::1] c, int sum_c, int n) nogil:
    cdef int i
    sum_c = c[0]
    for i in range(1, n):
        sum_c += c[i]

cpdef void dvec_minus_ivec(double[::1] c, double[::1] dvec, long[:] ivec, int n) nogil:
    cdef int i
    for i in range(n):
        c[i] = dvec[i]-ivec[i]

cpdef void elementwise_grid_parameters(double[::1] c, double[::1] c_min, double[::1] dc, long[::1] c_bins, double[::1] alpha, int sum_location, long[::1] location, double[::1] alpha_rest, double[::1] abs_alpha_rest, long[::1] d_location) nogil:
    cdef int n = alpha.shape[0]
    cdef int i
    #sum_location = 0
    for i in range(n):
        # Transform
        alpha[i] = (c[i] - c_min[i]) / dc[i]
        # Round
        location[i] = <int> alpha[i]
        if alpha[i] - location[i] > 0.5:
            location[i] += 1
        sum_location += location[i]
        alpha_rest[i] = alpha[i] - location[i]
        abs_alpha_rest[i] = fabs(alpha_rest[i])
        d_location[i] = 1 if alpha_rest[i] >= 0 else -1
        if alpha[i] < 0 or alpha[i] > c_bins[i]:
            inside = 0
            break

cpdef int c_argmax(double[::1] c, int max_ind, double c_max, int n) nogil:
    cdef int i
    max_ind = 0
    c_max = c[0]
    for i in range(1,n):
        if c_max < c[i]:
            c_max = c[i]
            max_ind = i
    return max_ind



#cython: language_level=3
import numpy as np
cimport numpy as npc
from .exp_and_log cimport *
from .indexing cimport *
from .caching cimport *
from .blas_functions cimport *
from scipy.linalg.cython_lapack cimport zheevd

# Unitary Interpolation
cdef class Hamiltonian_System:
    # Initialize variables, to quickly calculate interpolations while minimizing memmory allocation overheads
    cdef int n_dims, n_dims_1, d, d2, n_d_di_1, n_d_di,
    cdef long[::1] d_di
    cdef double[::1] E
    cdef double complex[::1] C1
    cdef double complex[:,:,::1] H
    cdef double complex[:,::1] V, U1, U2, U3, U4, U5
    cdef double complex *h0
    cdef double complex *v0
    cdef double complex *curr_h0
    cdef double complex *u1
    cdef double complex *u2
    cdef double complex *u3
    cdef double complex *u4
    cdef double complex *u5
    cdef double *e0
    cdef double complex *c1
    cdef int lwork, lrwork, liwork
    cdef double complex[::1] work
    cdef double complex *work0
    cdef double[::1] rwork
    cdef double *rwork0
    cdef int[::1] iwork
    cdef int *iwork0
    cdef char *jobz
    cdef char *uplo
    cdef int info
    def __cinit__(self, double complex[:,:,::1] H_s, long[::1] which_diffs = np.array([], dtype=long)):
        # Construct parameters
        self.n_dims = H_s.shape[0]
        self.n_dims_1 = self.n_dims - 1
        self.d = H_s.shape[1]
        self.d2 = self.d * self.d
        if which_diffs.shape[0] == 0:
            self.d_di = np.arange(self.n_dims_1)
        else:
            self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

        self.H = np.empty([self.n_dims, self.d, self.d], dtype=np.complex128)
        self.h0 = &self.H[0,0,0]
        cdef int i
        cdef double complex *h_s0 = &H_s[0,0,0]
        copy_pointer(h_s0, self.h0, self.d2*self.n_dims)

        self.E = np.empty([self.d], dtype=np.double)
        self.e0 = &self.E[0]
        self.C1 = np.empty([self.d], dtype=np.complex128)
        self.c1 = &self.C1[0]
        self.U1 = np.empty([self.d, self.d], dtype=np.complex128)
        self.u1 = &self.U1[0, 0]
        self.U2 = np.empty([self.d, self.d], dtype=np.complex128)
        self.u2 = &self.U2[0, 0]
        self.U3 = np.empty([self.d, self.d], dtype=np.complex128)
        self.u3 = &self.U3[0, 0]
        self.U4 = np.empty([self.d, self.d], dtype=np.complex128)
        self.u4 = &self.U4[0, 0]
        self.U5 = np.empty([self.d, self.d], dtype=np.complex128)
        self.u5 = &self.U5[0, 0]
        self.V = np.empty([self.d, self.d], dtype=np.complex128)
        self.v0 = &self.V[0, 0]
        self.curr_h0 = &self.H[0, 0, 0]

        self.lwork, self.lrwork, self.liwork = c_eigh_lapack_workspace_sizes(self.V)
        self.work = np.empty([self.lwork], dtype=np.complex128)
        self.work0 = &self.work[0]
        self.rwork = np.empty([self.lrwork], dtype=np.double)
        self.rwork0 = &self.rwork[0]
        self.iwork = np.empty([self.liwork], dtype=np.int32)
        self.iwork0 = &self.iwork[0]
        self.jobz = 'v'  #eigenvectors and values -> v
        self.uplo = 'l'  # upper triangle
        self.info = 0

    def set_which_diffs(self, long[::1] which_diffs):
        self.d_di = which_diffs
        self.n_d_di = self.d_di.shape[0]

    def get_which_diffs(self):
        return np.asarray(self.d_di), self.n_d_di

    cdef weighted_hamiltonian(self, double[::1] c):
        self.curr_h0 = self.h0
        copy_pointer(self.curr_h0, self.v0, self.d2)
        for i in range(self.n_dims_1):
            self.curr_h0 += self.d2
            d_mat_add_pointer(self.v0, self.curr_h0, c[i], self.d2)

    def expmH(self, double[::1] c, double complex[:,::1] U, double dt=1.0):
        # Construct Hamiltonian
        if not c.shape[0] == self.n_dims_1:
            raise ValueError('c.shape[0] needs to be equal to H_s[0].shape[0]-1.')
        cdef double complex *u0 = &U[0,0]
        self.weighted_hamiltonian(c)
        zheevd(self.jobz, self.uplo, &self.d, self.v0, &self.d, self.e0, self.work0, &self.lwork, self.rwork0, &self.lrwork, self.iwork0, &self.liwork, &self.info)
        copy_pointer(self.v0, self.u1, self.d2)
        v_exp_pointer(self.e0, self.u1, dt, self.d)
        DagM_M_cdot_pointer(self.v0, self.u1, u0, self.d)

    cdef dexpmH_pointer(self, double[::1] c, double complex *u0, double complex *du0, dt=1.0):  #int[::1] d_di,
        # Construct Hamiltonian
        cdef int i
        cdef double complex *h0
        cdef int curr_d_di_1
        self.weighted_hamiltonian(c)
        zheevd(self.jobz, self.uplo, &self.d, self.v0, &self.d, self.e0, self.work0, &self.lwork, self.rwork0, &self.lrwork, self.iwork0, &self.liwork, &self.info)
        copy_pointer(self.v0, self.u1, self.d2)
        v_exp_pointer(self.e0, self.u1, dt, self.d)
        DagM_M_cdot_pointer(self.v0, self.u1, u0, self.d)
        # Calculate the differentials
        for i in range(self.n_d_di): # According to https://arxiv.org/pdf/2006.00935.pdf (Daalgard, Motzoi paper --> Original source?)
            curr_d_di_1 = self.d_di[i] + 1
            h0 = &self.H[curr_d_di_1, 0, 0]
            MM_cdot_scale_pointer(dt, self.v0, h0, self.u1, self.d) #self.u1, self.d)
            M_DagM_cdot_pointer(self.u1, self.v0, du0, self.d)

            phase_shift_matrix_pointer(self.e0, self.c1, self.u1, dt, self.d)
            AxB_elementwise_pointer(du0, self.u1, self.u2, self.d2)

            MM_cdot_pointer( self.u2, self.v0, self.u1, self.d)
            DagM_M_cdot_pointer(self.v0, self.u1, du0, self.d)
            du0 += self.d2
    def dexpmH(self, double[::1] c, double complex[:,::1] U, double complex[:,:,::1] dU, double dt=1.0):  #int[::1] d_di,
        # d_di contains the indexes of the derivatives that we want to calculate (needs to be in ascending order with a negative value at the end)
        cdef double complex *u0 = &U[0, 0]
        cdef double complex *du0 = &dU[0,0,0]
        if not c.shape[0] == self.n_dims_1:
            raise ValueError('The coefficient c must be of size [interpolation_dimensions].')
        if not self.d_di.shape[0] == dU.shape[0]:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dU.shape[0].')
        self.dexpmH_pointer(c, u0, du0, dt)

    cdef expmH_pulse_pointer(self, double[:,::1] cs, double complex *u0, double dt=1.0):
        cdef int i
        cdef int steps = cs.shape[0]

        self.weighted_hamiltonian(cs[0,:])
        zheevd(self.jobz, self.uplo, &self.d, self.v0, &self.d, self.e0, self.work0, &self.lwork, self.rwork0, &self.lrwork, self.iwork0, &self.liwork, &self.info)
        copy_pointer(self.v0, self.u1, self.d2)
        v_exp_pointer(self.e0, self.v0, dt, self.d)
        DagM_M_cdot_pointer(self.u1, self.v0, u0, self.d)

        for i in range(1,steps):
            self.weighted_hamiltonian(cs[i, :])
            zheevd(self.jobz, self.uplo, &self.d, self.v0, &self.d, self.e0, self.work0, &self.lwork, self.rwork0, &self.lrwork, self.iwork0, &self.liwork, &self.info)
            MM_cdot_pointer(self.v0, u0, self.u1, self.d)
            v_exp_pointer(self.e0, self.u1, dt, self.d)
            DagM_M_cdot_pointer(self.v0, self.u1, u0, self.d)
    def expmH_pulse(self, double[:,::1] cs, double complex[:,::1] U, double dt=1.0):
        # Construct Hamiltonian
        cdef double complex *u0 = &U[0, 0]
        self.expmH_pulse_pointer(cs, u0, dt)

    def grape(self, double[:,::1] cs, double complex[:,::1] U_target, int[::1] target_indexes, double complex[:,::1] U, double complex[:,:,::1] dU, double[:,::1] dI_dj):
        # Calculate fidelity for a pulse and the differentials of the fidelity at every timestep using the grape trick
        cdef int i, j
        cdef int steps = cs.shape[0]

        cdef double complex *new_p0 = self.u1
        cdef double complex *new_q0 = self.u2
        cdef double complex *p0 = self.u4
        cdef double complex *q0 = self.u5
        cdef double complex *u0 = &U[0, 0]
        cdef double complex *u_tar = &U_target[0, 0]
        cdef double complex *curr_u = self.u3
        cdef double complex *du = &dU[0,0,0]

        cdef double complex trM, trdM, two_nni
        cdef double n = <double> target_indexes.shape[0]
        cdef double nni = 1.0 / (n * (n + 1))
        cdef double one_minus = 1.0 - n * nni

        if not dI_dj.shape[0] == steps:
            raise ValueError('Inputs must fulfill: cs.shape[0] = dI_dj.shape[0].')
        if not dI_dj.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dI_dj.shape[1].')
        if not cs.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = cs.shape[1].')
        # Fidelity constants

        self.expmH_pulse_pointer(cs, u0)
        #Dag_fast(U_target, self.Ur3) # -> Ur3 is identical to Q  -> replaced by new approach
        DagM_M_cdot_pointer(u_tar, u0, q0, self.d)
        trM = target_indexes_trace_pointer(q0, self.d, target_indexes)
        two_nni = - 2 / (n * (n + 1)) * conj(trM)
        cdef double I0 = one_minus - nni * abs_2(trM) # Sub Fidelity

        # Do GRAPE trick
        self.dexpmH_pointer(cs[0,:], curr_u, du) # Calculate Unitary and its derivatives
        M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Reduce q0
        copy_pointer(new_q0, q0, self.d2)
        for j in range(self.n_d_di):
            #MM_cdot_pointer(curr_du, p0, new_p0, self.d)  # p0 is not defined yet
            trdM = tr_dot_pointer_target_indexes(q0, du, self.d, target_indexes)
            dI_dj[0, j] = ( two_nni * trdM ).real
            du += self.d2
        copy_pointer(curr_u, p0, self.d2)  # Define p0 from curr_u
        for i in range(1, steps):
            du -= self.d2 * self.n_d_di
            self.dexpmH_pointer(cs[i, :], curr_u, du)  # Calculate Unitary and its derivatives
            M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Remove current unitary (curr_u) from Q
            for j in range(self.n_d_di):
                MM_cdot_pointer(du, p0, new_p0, self.d)
                trdM = tr_dot_pointer_target_indexes(new_q0, new_p0, self.d, target_indexes)
                dI_dj[i, j] = ( two_nni * trdM ).real
                du += self.d2
            MM_cdot_pointer(curr_u, p0, new_p0, self.d)  # Add current unitary (curr_u) to P
            # flip pointers
            self.u1, self.u4 = self.u4, self.u1
            self.u2, self.u5 = self.u5, self.u2
            new_p0, p0 = p0, new_p0
            new_q0, q0 = q0, new_q0
        return I0




#cython: language_level=3
import numpy as np
cimport numpy as npc
from libc.math cimport fabs
from .exp_and_log cimport *
from .indexing cimport *
from .caching cimport *
from .blas_functions cimport *


# Unitary Interpolation
cdef class UI:
    # Initialize variables, to quickly calculate interpolations while minimizing memmory allocation overheads
    cdef double[::1] c_mins, c_maxs, dcs, das
    cdef long[::1] c_bins
    cdef int n_dims, d, d2, n_dims_1, n_d_di_1, n_d_di
    cdef double[:, ::1] E
    cdef double complex[:,:,::1] Vr, Vl, CL, CH, dU1
    cdef double complex[:,::1] Ur, Ur1, Ur2, Ur3, Ur4, Ur5
    cdef double* ei
    cdef double complex *vl
    cdef double complex *vr
    cdef double complex *c
    cdef double complex *ur0
    cdef double complex *ur1
    cdef double complex *ur2
    cdef double complex *ur3
    cdef double complex *ur4
    cdef double complex *ur5
    cdef double complex *du1
    cdef long[::1] strides_L, strides_R,
    cdef long[:,::1] strides_E, strides_C
    cdef long[::1] location, d_location
    cdef double[::1] abs_alpha_rest, alpha
    cdef long[::1] first_elements_E, first_elements_C, L
    cdef long[::1] d_di
    def __cinit__(self, double complex[:,:,::1] H_s, double[::1] c_min_s, double[::1] c_max_s, long[::1] c_bins, long[::1] which_diffs = np.array([], dtype=long)):
        # Construct parameters
        self.n_dims = c_min_s.shape[0]
        self.n_dims_1 = self.n_dims - 1
        self.d = H_s.shape[1]
        self.d2 = self.d * self.d
        if not H_s.shape[0] == self.n_dims + 1:
            print('Requires n+1 Hamiltonians for n dimensional interpolation. Check lenths of Hs, c_mins, c_maxs, c_bins')
            raise ValueError
        self.c_bins = np.empty(self.n_dims, dtype=long)
        for i in range(self.n_dims):
            self.c_bins[i] = c_bins[i]
        self.c_mins, self.c_maxs, self.dcs, self.das = Bin_Parameters(c_min_s, c_max_s, self.c_bins)
        # Single Step Indexing Parameters  --> Add multiple indexes later
        self.location = np.empty(self.n_dims, dtype=long)
        self.abs_alpha_rest = np.empty(self.n_dims, dtype=np.double)
        self.d_location = np.empty(self.n_dims, dtype=long)
        self.alpha = np.empty(self.n_dims, dtype=np.double)
        if which_diffs.shape[0] == 0:
            self.d_di = np.arange(self.n_dims)
        else:
            self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

        # Construct interpolation grid points
        U_grid, cum_prod = Unitary_Grid(H_s, self.c_mins, self.dcs, self.c_bins)
        ## Construct interpolation cache
        self.E, self.Vl, self.Vr, self.CL, self.CH, self.strides_E, self.strides_L, self.strides_R, self.strides_C, self.first_elements_E, self.first_elements_C = Create_Interpolation_Cache( U_grid, cum_prod, self.c_bins)
        self.ei = &self.E[0, 0]
        self.vl = &self.Vl[0, 0, 0]
        self.vr = &self.Vr[0, 0, 0]
        self.c = &self.CH[0, 0, 0]
        self.Ur = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur0 = &self.Ur[0, 0]
        self.Ur1 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur1 = &self.Ur1[0, 0]
        self.Ur2 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur2 = &self.Ur2[0, 0]
        self.Ur3 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur3 = &self.Ur3[0, 0]
        self.Ur4 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur4 = &self.Ur4[0, 0]
        self.Ur5 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur5 = &self.Ur5[0, 0]
        self.dU1 = np.empty([self.n_dims+1, self.d, self.d], dtype=np.complex128)
        self.du1 = &self.dU1[0, 0, 0]
        self.L = np.empty(self.n_dims, dtype=long)

    cdef single_parameters2oddgrid(self, double[::1] c):
        cdef int sum_location = 0
        cdef double alpha_max = 0.0
        cdef double alpha_rest
        cdef int i
        cdef int max_alpha_ind = 0
        cdef int max_vals
        for i in range(self.n_dims):
            # Transform
            self.alpha[i] = (c[i] - self.c_mins[i]) / self.dcs[i]
            # Round
            self.location[i] = <int> self.alpha[i]     # floor -> 
            alpha_rest = self.alpha[i] - self.location[i]
            if alpha_rest > 0.5:
                self.location[i] += 1
                alpha_rest -= 1
            sum_location += self.location[i]

            self.abs_alpha_rest[i] = fabs(alpha_rest)
            self.d_location[i] = 1 if alpha_rest >= 0 else -1
            if self.alpha[i] < 0 or self.alpha[i] > self.c_bins[i]:
                print('Warning: These parameters lie outside of interpolation grid!')
                break
        if is_even(sum_location):
            max_alpha_ind = c_argmax(self.abs_alpha_rest, max_alpha_ind, alpha_max, self.n_dims)
            if self.location[max_alpha_ind] + self.d_location[max_alpha_ind] > self.c_bins[max_alpha_ind]:
                self.d_location[max_alpha_ind] = -1
            self.location[max_alpha_ind] += self.d_location[max_alpha_ind]
            self.d_location[max_alpha_ind] = - self.d_location[max_alpha_ind]
            self.abs_alpha_rest[max_alpha_ind] = 1-self.abs_alpha_rest[max_alpha_ind]
        for i in range(self.n_dims):
            self.d_location[i] -= 1
            self.d_location[i] >>= 1 # via bitshift operation
            #d_location[i] /= 2 #(d_location[i]-1)/2 ### in two steps with optimized in-place operations
        for i in range(self.n_dims):
            max_vals = self.location[i]+self.d_location[i]
            if max_vals > self.c_bins[i]-1:
                self.d_location[i] = -1

    def set_which_diffs(self, long[::1] which_diffs):
        self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

    def get_single_param(self, double[::1] c): # To verify interpolation grid
        self.single_parameters2oddgrid(c)
        return np.asarray(self.location), np.asarray(self.d_location), np.asarray(self.abs_alpha_rest)

    def get_cache(self): # To verify interpolation grid
        return np.asarray(self.E), np.asarray(self.Vl), np.asarray(self.Vr), np.asarray(self.CL), np.asarray(self.CH), np.asarray(self.strides_E), np.asarray(self.strides_L), np.asarray(self.strides_R), np.asarray(self.strides_C)

    cdef interpolate_single_u(self, double complex *u0): #u0 => input the matrices for output
        cdef int i, j
        cdef int ind
        if self.n_dims == 1:
            ind = self.location[0] + self.d_location[0]
            self.vr = &self.Vr[ind, 0, 0] # In 1D strides are 1
            self.vl = &self.Vl[ind,0,0]
            self.ei = &self.E[ind,0]
            copy_pointer(self.vr, self.ur0, self.d2)
            v_exp_pointer(self.ei, self.ur0, self.abs_alpha_rest[0], self.d)
            MM_cdot_pointer(self.vl, self.ur0, u0, self.d)
        else:
            # Right side first
            self.L[0] = self.location[0] + self.d_location[0]
            for i in range(1, self.n_dims):
                self.L[i] = self.location[i]
            ind = findex_0(self.L, self.strides_E[0,:], self.n_dims)
            self.vr = &self.Vr[ind, 0, 0]  # In 1D strides are 1
            copy_pointer(self.vr, self.ur0, self.d2)
            self.ei = &self.E[ind, 0]
            v_exp_pointer(self.ei, self.ur0, self.abs_alpha_rest[0], self.d)
            self.L[0] = self.location[0]
            # Center multiplications
            for i in range(self.n_dims_1):
                j = i + 1
                self.L[j] += self.d_location[j]
                ind = findex_0(self.L, self.strides_E[j, :], self.n_dims) + self.first_elements_E[j]
                self.ei = &self.E[ind, 0]
                if self.d_location[i]:  ### Higher
                    self.L[i] += -1
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.c = &self.CL[ind, 0, 0]
                    self.L[i] += 1  # Restore value
                else:                        ### Lower
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims)+self.first_elements_C[i]
                    self.c = &self.CH[ind,0,0]
                self.L[j] = self.location[j]
                MM_cdot_pointer(self.c, self.ur0, self.ur1, self.d)
                v_exp_pointer(self.ei, self.ur1, self.abs_alpha_rest[j], self.d)
                self.ur1, self.ur0 = self.ur0, self.ur1
            # Left side multiplication
            self.L[self.n_dims_1] += self.d_location[self.n_dims_1]
            ind = findex_0(self.L, self.strides_L, self.n_dims)
            self.vl = &self.Vl[ind,0,0]
            MM_cdot_pointer(self.vl, self.ur0, u0, self.d)

    cdef interpolate_single_u_du(self, double complex *u0, double complex *du0): #u0, du0 => input the matrices for output   ##int[::1] d_di, -> define earlier
        # d_di contains the indexes of the derivatives that we want to calculate
        cdef int i, j, ind, curr_d_di
        cdef int curr_d_ind = 0
        cdef int do_copy = 0
        cdef double complex *du1 = self.du1 #.copy()
        cdef double complex *du1_0 = self.du1 #.copy()
        if self.n_dims == 1:
            ind = self.location[0] + self.d_location[0]
            self.vr = &self.Vr[ind, 0, 0] # In 1D strides are 1
            self.vl = &self.Vl[ind, 0, 0]
            self.ei = &self.E[ind, 0]
            copy_pointer(self.vr, self.ur0, self.d2)
            if self.d_di[curr_d_ind] == 0:
                copy_pointer(self.vr, du1, self.d2)
                v_exp_v_pointer( (self.d_location[0]*2+1)*self.das[0], self.ei, du1, self.abs_alpha_rest[0], self.d) # -i*E*amp*exp(-i*E*t)
                MM_cdot_pointer(self.vl, du1, du0, self.d)
            v_exp_pointer(self.ei, self.ur0, self.abs_alpha_rest[0], self.d)
            MM_cdot_pointer(self.vl, self.ur0, u0, self.d)
        else:
            # Right side first
            self.L[0] = self.location[0] + self.d_location[0]
            for i in range(1, self.n_dims):
                self.L[i] = self.location[i]
            ind = findex_0(self.L, self.strides_E[0,:], self.n_dims)
            self.vr = &self.Vr[ind, 0, 0]  # In 1D strides are 1
            self.ei = &self.E[ind, 0]
            copy_pointer(self.vr, du1, self.d2)
            curr_d_di = self.d_di[curr_d_ind]
            if curr_d_di == 0:
                copy_pointer(du1, du0, self.d2)
                v_exp_v_pointer( (self.d_location[curr_d_di]*2+1)*self.das[curr_d_di], self.ei, du0, self.abs_alpha_rest[0], self.d)  # -i*E*amp*exp(-i*E*t)
                curr_d_ind += 1
                curr_d_di = self.d_di[curr_d_ind]
                du1 += self.d2 # Shift pointers appropriately by one matrix
                du0 += self.d2
                do_copy = 1
            v_exp_pointer(self.ei, du1_0, self.abs_alpha_rest[0], self.d)
            self.L[0] = self.location[0]
            # Center multiplications
            for i in range(self.n_dims_1):
                j = i + 1
                self.L[j] += self.d_location[j]
                ind = findex_0(self.L, self.strides_E[j, :], self.n_dims) + self.first_elements_E[j]
                self.ei = &self.E[ind, 0]
                if self.d_location[i]:  ### Higher
                    self.L[i] += -1
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.c = &self.CL[ind, 0, 0]
                    self.L[i] += 1  # Restore value
                else:                        ### Lower
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims)+self.first_elements_C[i]
                    self.c = &self.CH[ind,0,0]
                self.L[j] = self.location[j]
                if do_copy:
                    do_copy = 0
                    copy_pointer(self.c, self.ur1, self.d2)
                else:
                    MM_cdot_pointer(self.c, self.ur0, self.ur1, self.d)
                if curr_d_di == j: # Add derivative here
                    # Multiply onto du1_0 element -> for final unitary
                    MM_cdot_pointer(self.ur1, du1_0, self.ur0, self.d)
                    copy_pointer(self.ur0, du1_0, self.d2)
                    # Multiply onto curr du element (via du0 pointer) --> Change derivative term
                    copy_pointer(du1_0, du0, self.d2)
                    v_exp_v_pointer( (self.d_location[curr_d_di]*2+1)*self.das[curr_d_di], self.ei, du0, self.abs_alpha_rest[j], self.d)  # -i*E*amp*exp(-i*E*t)
                    # Change unitary element
                    v_exp_pointer(self.ei, du1_0, self.abs_alpha_rest[j], self.d)
                    # Multiply onto partial storage matrix du1
                    if curr_d_ind > 0:
                        copy_pointer(self.ur1, du1, self.d2)
                        v_exp_pointer(self.ei, du1, self.abs_alpha_rest[j], self.d)
                    curr_d_ind += 1
                    curr_d_di = self.d_di[curr_d_ind]
                    du1 += self.d2  # Shift pointers appropriately by one matrix
                    du0 += self.d2
                    do_copy = 1
                else:
                    v_exp_pointer(self.ei, self.ur1, self.abs_alpha_rest[j], self.d)
                self.ur1, self.ur0 = self.ur0, self.ur1
            # Left side multiplication
            self.L[self.n_dims_1] = self.location[self.n_dims_1] + self.d_location[self.n_dims_1]
            ind = findex_0(self.L, self.strides_L, self.n_dims)
            self.vl = &self.Vl[ind,0,0]
            if do_copy:
                # do_copy = 0 # No longer necessary
                copy_pointer(self.vl, self.ur1, self.d2)
            else:
                MM_cdot_pointer(self.vl, self.ur0, self.ur1, self.d)
            MM_cdot_pointer(self.ur1, du1_0, u0, self.d)  # Finalize unitary and store on output matrix
            # Finalize by grouping the intermediate results
            #if curr_d_ind > 0:
            #    copy_pointer(self.ur1, du1, self.d2)
            du0 -= self.d2
            MM_cdot_pointer(self.ur1, du0, self.ur2, self.d)
            copy_pointer(self.ur2, du0, self.d2)
            for i in range(self.n_d_di_1):
                du0 -= self.d2
                du1 -= self.d2
                MM_cdot_pointer(self.ur1, du1, self.ur0, self.d)
                MM_cdot_pointer(self.ur0, du0, self.ur2, self.d)
                copy_pointer(self.ur2, du0, self.d2)
                self.ur1, self.ur0 = self.ur0, self.ur1

    cdef expmH_pointer(self, double[::1] c, double complex *u0):
        if not c.shape[0] == self.n_dims:
            raise ValueError('The coefficient c must be of size [interpolation_dimensions].')
        self.single_parameters2oddgrid(c)
        self.interpolate_single_u(u0)
    def expmH(self, double[::1] c, double complex[:,::1] U):
        cdef double complex *u0 = &U[0, 0]
        self.expmH_pointer(c, u0)

    cdef dexpmH_pointer(self, double[::1] c, double complex *u0, double complex *du0):  #int[::1] d_di,
        # d_di contains the indexes of the derivatives that we want to calculate (needs to be in ascending order with a negative value at the end)
        self.single_parameters2oddgrid(c)
        self.interpolate_single_u_du(u0, du0)

    def dexpmH(self, double[::1] c, double complex[:,::1] U, double complex[:,:,::1] dU):  #int[::1] d_di,
        # d_di contains the indexes of the derivatives that we want to calculate (needs to be in ascending order with a negative value at the end)
        cdef double complex *u0 = &U[0, 0]
        cdef double complex *du0 = &dU[0,0,0]
        if not c.shape[0] == self.n_dims:
            raise ValueError('The coefficient c must be of size [interpolation_dimensions].')
        if not self.d_di.shape[0] == dU.shape[0]:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dU.shape[0].')
        self.dexpmH_pointer(c, u0, du0)

    cdef expmH_pulse_pointer(self, double[:,::1] cs, double complex *u0):
        cdef int i = 0
        cdef int i_1
        cdef int steps = cs.shape[0]
        cdef int steps_1 = steps - 1
        cdef double complex *ur2 = self.ur2
        cdef double complex *ur3 = self.ur3
        if not cs.shape[1] == self.n_dims:
            raise ValueError('The coefficient matrix must be of size [n_time_steps, interpolation_dimensions].')
        # First timestep
        if steps_1 % 2:
            ur3, u0 = u0, ur3
        self.single_parameters2oddgrid(cs[i,:])
        self.interpolate_single_u(u0)
        # Central timesteps
        for i in range(1, steps_1, 2):
            self.single_parameters2oddgrid(cs[i,:])
            self.interpolate_single_u(ur2)
            MM_cdot_pointer(ur2, u0, ur3, self.d)
            i_1 = i + 1
            self.single_parameters2oddgrid(cs[i_1,:])
            self.interpolate_single_u(ur2)
            MM_cdot_pointer(ur2, ur3, u0, self.d)
        # Final timestep (if necessary)ts apo
        if steps_1 % 2:
            self.single_parameters2oddgrid(cs[steps_1,:])
            self.interpolate_single_u(ur2)
            MM_cdot_pointer(ur2, u0, ur3, self.d)
    def expmH_pulse(self, double[:,::1] cs, double complex[:,::1] U):
        cdef double complex *u0 = &U[0, 0]
        self.expmH_pulse_pointer(cs, u0)

    def grape(self, double[:,::1] cs, double complex[:,::1] U_target, int[::1] target_indexes, double complex[:,::1] U, double complex[:,:,::1] dU, double[:,::1] dI_dj):
        # Calculate fidelity for a pulse and the differentials of the fidelity at every timestep using the grape trick
        cdef int i, j
        cdef int steps = cs.shape[0]

        cdef double complex *new_p0 = self.ur0
        cdef double complex *new_q0 = self.ur1
        cdef double complex *p0 = self.ur4
        cdef double complex *q0 = self.ur5
        cdef double complex *u0 = &U[0, 0]
        cdef double complex *u_tar = &U_target[0, 0]
        cdef double complex *curr_u = self.ur3
        cdef double complex *du = &dU[0,0,0]

        cdef double complex trM, trdM, two_nni
        cdef double n = <double> target_indexes.shape[0]
        cdef double nni = 1.0 / (n * (n + 1))
        cdef double one_minus = 1.0 - n * nni

        if not dI_dj.shape[0] == steps:
            raise ValueError('Inputs must fulfill: cs.shape[0] = dI_dj.shape[0].')
        if not dI_dj.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = dI_dj.shape[1].')
        if not cs.shape[1] == self.n_d_di:
            raise ValueError('Inputs must fulfill: which_diffs.shape[0] = cs.shape[1].')
        # Fidelity constants

        self.expmH_pulse_pointer(cs, u0)
        #Dag_fast(U_target, self.Ur3) # -> Ur3 is identical to Q  -> replaced by new approach
        DagM_M_cdot_pointer(u_tar, u0, q0, self.d)
        trM = target_indexes_trace_pointer(q0, self.d, target_indexes)
        two_nni = - 2 / (n * (n + 1)) * conj(trM)
        cdef double I0 = one_minus - nni * abs_2(trM) # Sub Fidelity

        # Do GRAPE trick
        self.dexpmH_pointer(cs[0,:], curr_u, du) # Calculate Unitary and its derivatives
        M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Reduce q0
        copy_pointer(new_q0, q0, self.d2)
        for j in range(self.n_d_di):
            #MM_cdot_pointer(curr_du, p0, new_p0, self.d)  # p0 is not defined yet
            trdM = tr_dot_pointer_target_indexes(q0, du, self.d, target_indexes)
            dI_dj[0, j] = ( two_nni * trdM ).real
            du += self.d2
        copy_pointer(curr_u, p0, self.d2)  # Define p0 from curr_u
        for i in range(1, steps):
            du -= self.d2 * self.n_d_di
            self.dexpmH_pointer(cs[i, :], curr_u, du)  # Calculate Unitary and its derivatives
            M_DagM_cdot_pointer(q0, curr_u, new_q0, self.d)  # Remove current unitary (curr_u) from Q
            for j in range(self.n_d_di):
                MM_cdot_pointer(du, p0, new_p0, self.d)
                trdM = tr_dot_pointer_target_indexes(new_q0, new_p0, self.d, target_indexes)
                dI_dj[i, j] = ( two_nni * trdM ).real
                du += self.d2
            MM_cdot_pointer(curr_u, p0, new_p0, self.d)  # Add current unitary (curr_u) to P
            #copy_pointer(new_p0, p0, self.d2)
            #copy_pointer(new_q0, q0, self.d2)
            # flip pointers
            self.ur0, self.ur4 = self.ur4, self.ur0
            self.ur1, self.ur5 = self.ur5, self.ur1
            new_p0, p0 = p0, new_p0
            new_q0, q0 = q0, new_q0
        return I0




#cython: language_level=3
from __future__ import print_function
cimport cython
import numpy as np
from .blas_functions cimport c_eigh_lapack_workspace_sizes, d_mat_add, M_DagM_cdot, MM_cdot, DagM_M_cdot
from .exp_and_log cimport c_expmH, Make_U_Partial, Dag, Partialize_Center, Partialize_Sides, copy_pointer  # Find correct files for these functions
from .indexing cimport next_in_mrange, findex_0, reverse_cum_prod, flat_index_sizes_E, flat_index_sizes_V, int_max, int_prod_array, int_sum, is_even         # Find correct files for these functions


##### Create Interpolation Cache #######################################################################################
cpdef Unitary_Grid(double complex[:,:,::1] H, double[::1] c_mins, double[::1] dcs, long[::1] c_bins): # s is the number of interpolation points
    cdef int n = H.shape[1]
    cdef int l = c_bins.shape[0]
    if not H.shape[0]  == l + 1:
        print('Error: The number of Hamiltonians needs to be one higher than the number of control parameters')
        raise ValueError
    cdef long[:] d_bins = np.empty(l, dtype=long)
    cdef int i, j # int
    for i in range(l):
        d_bins[i] = c_bins[i] + 1
    cdef long[:] cum_prod = np.empty(l, dtype=long)
    cdef int tot_prod
    cum_prod, tot_prod = reverse_cum_prod(d_bins, l)
    cdef double complex[:,:,::1] U_grid = np.empty([tot_prod, n, n], dtype=np.complex128)
    cdef double complex[:,::1] H_i_s = np.empty([n, n], dtype=np.complex128)
    cdef long[:] i_s = np.zeros(l, dtype=long)
    cdef double d
    cdef double dt = 1.0
    work, rwork, iwork = c_eigh_lapack_workspace_sizes(H_i_s)
    for i in range(tot_prod):
        H_i_s[:,:] = H[0, :, :]
        for j in range(l):
            d = (c_mins[j] + dcs[j] * i_s[j])
            d_mat_add(H_i_s, H[<int> j+1,:,:], d)
        #print(str(findex_0(i_s, cum_prod, l)))
        c_expmH(H_i_s, dt, U_grid[findex_0(i_s, cum_prod, l), :, :], work, rwork, iwork)
        next_in_mrange(i_s, d_bins) # returns new i_s -> always apply at the end of the loop
    return U_grid, cum_prod


#### Needs to be updates
cpdef Create_Interpolation_Cache(double complex[:,:,::1] U_grid, long[::1] grid_cum_prod, long[::1] c_bins):
    # Create interpolation arrows
    cdef int n = U_grid.shape[1]
    cdef int l = c_bins.shape[0]
    cdef int ind, i_1
    cdef long[::1] i_s = np.empty(l, dtype = long)
    cdef long[::1] i_s2 = np.empty(l, dtype = long)
    cdef long[::1] j_s = np.empty(l, dtype = long)
    cdef long[::1] k_s = np.empty(l, dtype = long)
    cdef long[::1] curr_arrow_index = np.empty(l, dtype = long)
    cdef int i, j  # int
    cdef int tot_ind_E, tot_ind_L, tot_ind_C, tot_ind_R
    cdef long[::1] first_elements_E = np.empty(l, dtype = long)
    cdef long[::1] first_elements_C = np.empty(l-1, dtype = long)

    cdef long[:,::1] array_strides_E = np.empty([l,l], dtype = long)
    cdef long[::1] array_strides_L = np.empty(l, dtype = long)
    cdef long[::1] array_strides_R = np.empty(l, dtype = long)
    cdef long[:,::1] array_strides_C = np.empty([l-1,l], dtype = long)

    cdef long[:,::1] array_sizes_E = np.empty([l,l], dtype = long)
    cdef long[::1] array_sizes_L = np.empty(l, dtype = long)
    cdef long[::1] array_sizes_R = np.empty(l, dtype = long)
    cdef long[:,::1] array_sizes_C = np.empty([l-1,l], dtype = long)

    tot_ind_E, first_elements_E, array_strides_E, array_sizes_E = flat_index_sizes_E(c_bins)
    tot_ind_L, array_strides_L, array_sizes_L, tot_ind_C, first_elements_C, array_strides_C, array_sizes_C, tot_ind_R, array_strides_R, array_sizes_R = flat_index_sizes_V(c_bins)
    # Initialize cache matrices
    cdef double[:, ::1] E = np.empty([tot_ind_E, n], dtype = np.double)
    cdef double complex[:,:,::1] Vr = np.empty([tot_ind_R, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] Vl = np.empty([tot_ind_L, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CL = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CH = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    # Initialize Intermediate variables
    cdef double complex[:,::1] U_0     = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] U_1     = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] V       = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] logU    = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] V_tmp   = np.empty([n,n], dtype = np.complex128)
    cdef double [::1]         logE_tmp = np.empty(n, dtype = np.double)
    cdef double[::1]                Es = np.empty(n, dtype = np.double)
    cdef double complex[::1]        Ec = np.empty(n, dtype = np.complex128)
    cdef double complex[:,::1]    Emat = np.empty([n,n], dtype = np.complex128)
    cdef long[::1] strides       = np.empty(l, dtype = long)
    cdef long[::1] prev_strides  = np.empty(l, dtype = long)
    cdef long[::1] strides_V     = np.empty(l, dtype = long)
    cdef double complex[:,::1] L = np.empty([n,n], dtype = np.complex128)
    cdef long[::1] sub_mat_length = int_prod_array(array_sizes_E, l)
    cdef int max_tot_ind = int_max(sub_mat_length, l)
    cdef int prod_array_sizes
    cdef double complex[:,:,::1] curr_V = np.empty([max_tot_ind,n,n], dtype = np.complex128)
    cdef double complex[:,:,::1] prev_V = np.empty([max_tot_ind,n,n], dtype = np.complex128)
    # Get eig parameters
    for i in range(l): # One arrow cache for every direction in grid
        sizes = array_sizes_E[i,:]
        strides = array_strides_E[i,:]
        i_1 = i - 1
        if i > 0:
            prev_strides[:] = array_strides_E[i_1, :]
            strides_V[:] = array_strides_C[i_1, :]
        i_s = np.zeros(l, dtype = long) # Initialize as zero -> update at end of loop via next_in_mrange
        prod_array_sizes = sub_mat_length[i]
        for j in range(prod_array_sizes):
            for k in range(l):
                if not k == i:
                    j_s[k] = i_s[k]
                else:
                    j_s[k] = i_s[k]+1
            if is_even(int_sum(i_s, l)):
                curr_arrow_index = j_s
                U_1 = U_grid[findex_0(i_s, grid_cum_prod, l), :, :]
                U_0 = U_grid[findex_0(j_s, grid_cum_prod, l), :, :]
            else:
                curr_arrow_index = i_s
                U_0 = U_grid[findex_0(i_s, grid_cum_prod, l), :, :]
                U_1 = U_grid[findex_0(j_s, grid_cum_prod, l), :, :]
            Make_U_Partial(U_1, U_0, Emat, Ec, logE_tmp, V_tmp, Es, V) # returns E, V
            curr_V[findex_0(i_s, strides, l), :, :] = V
            ind = findex_0(i_s, strides, l) + first_elements_E[i]
            E[ind, :] = Es

            if i == 0: # Add right side only for i == 0
                MM_cdot(V, U_0, Emat)
                Vr[findex_0(i_s, strides, l),:,:] = Emat
            else:
                if i_s[i_1] > 0: #Lowering is possible
                    for k in range(l):
                        if not k == i-1:
                            k_s[k] = curr_arrow_index[k]
                            i_s2[k] = i_s[k]
                        else:
                            k_s[k] = curr_arrow_index[k] - 1
                            i_s2[k] = i_s[k] - 1
                    i_s2 = i_s.copy()
                    i_s2[i_1] += -1
                    M_DagM_cdot(V, prev_V[findex_0(k_s, prev_strides, l), :, :], Emat)
                    ind = findex_0(i_s2, strides_V, l) + first_elements_C[i_1]
                    CL[ind,:,:] =  Emat
                if i_s[i_1] < c_bins[i_1]: # Increasing is possible
                    M_DagM_cdot(V, prev_V[findex_0(curr_arrow_index, prev_strides,l), :, :], Emat)
                    ind = findex_0(i_s, strides_V, l) + first_elements_C[i_1]
                    CH[ind,:,:] = Emat      #dot(Dag(V), prev_VH)
            if i == l-1:
                L = Dag(V)
                Vl[findex_0(i_s, strides, l),:,:] = L #Dag(V)
            next_in_mrange(i_s, sizes)   # Increase i_s counter
        prev_V[:,:,::1] = curr_V
    return E, Vl, Vr, CL, CH, array_strides_E, array_strides_L, array_strides_R, array_strides_C, first_elements_E, first_elements_C


#### Symmetric Unitary Interpolation
cpdef Create_Sym_Interpolation_Cache(double complex[:,:,::1] U_grid2, long[::1] grid_cum_prod, long[::1] c_bins):
    # U_grid2 = sqrt(U_grid)   -> for half steps
    # Create interpolation arrows
    U_grid = 1
    cdef int n = U_grid2.shape[1]
    cdef int l = c_bins.shape[0]
    cdef int n2 = n*n
    cdef int ind, i_1
    cdef long[::1] i_s = np.empty(l, dtype = long)
    cdef long[::1] i_s2 = np.empty(l, dtype = long)
    cdef long[::1] j_s = np.empty(l, dtype = long)
    cdef long[::1] k_s = np.empty(l, dtype = long)
    cdef long[::1] curr_arrow_index = np.empty(l, dtype = long)
    cdef int i, j  # int
    cdef int tot_ind_E, tot_ind_C, tot_ind_L
    cdef long[::1] first_elements_E = np.empty(l, dtype = long)
    cdef long[::1] first_elements_C = np.empty(l-1, dtype = long)

    cdef long[:,::1] array_strides_E = np.empty([l,l], dtype = long)
    cdef long[::1] array_strides_L   = np.empty(l, dtype = long)  # Also valid for L
    cdef long[:,::1] array_strides_C = np.empty([l-1,l], dtype = long)

    cdef long[:,::1] array_sizes_E = np.empty([l,l], dtype = long)
    cdef long[::1] array_sizes_L   = np.empty(l, dtype = long)    # Also valid for L
    cdef long[:,::1] array_sizes_C = np.empty([l-1,l], dtype = long)

    tot_ind_E, first_elements_E, array_strides_E, array_sizes_E = flat_index_sizes_E(c_bins)
    tot_ind_L, array_strides_L, array_sizes_L, tot_ind_C, first_elements_C, array_strides_C, array_sizes_C, _, _, _ = flat_index_sizes_V(c_bins)

    # Initialize cache matrices
    cdef double[:, ::1] E   = np.empty([tot_ind_E, n], dtype = np.double)
    cdef double complex[:,:,::1] Vr      = np.empty([tot_ind_L, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] Vl      = np.empty([tot_ind_L, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CL_R    = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CH_R    = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CL_L    = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    cdef double complex[:,:,::1] CH_L    = np.empty([tot_ind_C, n, n], dtype = np.complex128)
    # Initialize Intermediate variables
    cdef double complex[:,::1] U_0       = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] U_1       = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1] V_L       = np.empty([n,n], dtype = np.complex128)
    cdef double complex            *v_l0 = &V_L[0, 0]
    cdef double complex[:,::1] V_R       = np.empty([n,n], dtype = np.complex128)
    cdef double complex            *v_r0 = &V_R[0, 0]
    cdef double complex[:,::1] logU      = np.empty([n,n], dtype = np.complex128)
    cdef double complex[:,::1]   V_tmp   = np.empty([n,n], dtype = np.complex128)
    cdef double [::1]           logE_tmp = np.empty(n, dtype = np.double)
    cdef double[::1]                Es   = np.empty([n], dtype = np.double)
    cdef double complex[::1]        Ec   = np.empty([n], dtype = np.complex128)    # Intermediate variable
    cdef double complex[:,::1]    Emat   = np.empty([n,n], dtype = np.complex128)  # Intermediate variable
    #cdef double complex[:,::1]    U1     = np.empty([n,n], dtype = np.complex128)  # Intermediate variable
    cdef long[::1] strides        = np.empty(l, dtype = long)
    cdef long[::1] prev_strides   = np.empty(l, dtype = long)
    cdef long[::1] strides_V      = np.empty(l, dtype = long)
    cdef long[::1] sub_mat_length = int_prod_array(array_sizes_E, l)
    cdef int max_tot_ind         = int_max(sub_mat_length, l)

    cdef int prod_array_sizes
    cdef double complex[:,:,::1] curr_V_L  = np.empty([max_tot_ind,n,n], dtype = np.complex128)
    cdef double complex[:,:,::1] prev_V_L  = np.empty([max_tot_ind,n,n], dtype = np.complex128)
    cdef double complex[:,:,::1] curr_V_R  = np.empty([max_tot_ind,n,n], dtype = np.complex128)
    cdef double complex[:,:,::1] prev_V_R  = np.empty([max_tot_ind,n,n], dtype = np.complex128)

    for i in range(l): # One arrow cache for every direction in grid
        sizes = array_sizes_E[i, :]
        strides = array_strides_E[i, :]
        i_1 = i - 1
        if i > 0:
            prev_strides[:] = array_strides_E[i_1, :]
            strides_V[:] = array_strides_C[i_1, :]
        i_s = np.zeros(l, dtype = long) # Initialize as zero -> update at end of loop via next_in_mrange
        prod_array_sizes = sub_mat_length[i]

        for j in range(prod_array_sizes):
            for k in range(l):
                if not k == i:
                    j_s[k] = i_s[k]
                else:
                    j_s[k] = i_s[k]+1
            # Select matrices from grid (half step matrices)
            if is_even(int_sum(i_s, l)):
                curr_arrow_index = j_s
                U_1 = U_grid2[findex_0(i_s, grid_cum_prod, l), :, :]
                U_0 = U_grid2[findex_0(j_s, grid_cum_prod, l), :, :]
            else:
                curr_arrow_index = i_s
                U_0 = U_grid2[findex_0(i_s, grid_cum_prod, l), :, :]
                U_1 = U_grid2[findex_0(j_s, grid_cum_prod, l), :, :]

            if i == 0: # Central element (sqrt(U0) @ U1 @ sqrt(U0)) -> sqrt disappears in formulas as we are passing half steps to this function in the first place
                Partialize_Center(U_1, U_0, logU, Emat, Ec, logE_tmp, V_tmp, Es, V_L)  # returns E, V from decomposition of product: Dag(U_0) @ U_1^2 @ Dag(U_0)
                copy_pointer(v_l0, v_r0, n2)
            else:  # Calculate left and right hand side of this
                Partialize_Sides(U_1, U_0, Emat, Ec, logE_tmp, V_tmp, Es, V_L, V_R)
            ind = findex_0(i_s, strides, l) + first_elements_E[i]
            E[ind, :] = Es
            curr_V_L[findex_0(i_s, strides, l), :, :] = V_L
            curr_V_R[findex_0(i_s, strides, l), :, :] = V_R

            if i > 0:
                if i_s[i_1] > 0: # Lowering is possible
                    for k in range(l):
                        if not k == i-1:
                            k_s[k] = curr_arrow_index[k]
                            i_s2[k] = i_s[k]
                        else:
                            k_s[k] = curr_arrow_index[k] - 1
                            i_s2[k] = i_s[k] - 1
                    i_s2 = i_s.copy()
                    i_s2[i_1] += -1
                    ind = findex_0(i_s2, strides_V, l) + first_elements_C[i_1]
                    M_DagM_cdot(V_L, prev_V_L[findex_0(k_s, prev_strides, l), :, :], Emat)
                    CL_L[ind, :, :] = Emat
                    M_DagM_cdot(prev_V_R[findex_0(k_s, prev_strides, l), :, :], V_R, Emat)
                    CL_R[ind,:,:] =  Emat
                if i_s[i_1] < c_bins[i_1]: # Increasing is possible
                    ind = findex_0(i_s, strides_V, l) + first_elements_C[i_1]
                    M_DagM_cdot(V_L, prev_V_L[findex_0(curr_arrow_index, prev_strides, l), :, :], Emat)
                    CH_L[ind, :, :] = Emat
                    M_DagM_cdot(prev_V_R[findex_0(curr_arrow_index, prev_strides, l), :, :], V_R, Emat)
                    CH_R[ind, :, :] = Emat
            
            if i == l-1:
                M_DagM_cdot(U_0, V_L, Emat)
                Vl[findex_0(i_s, strides, l), :, :] = Emat
                MM_cdot(V_R, U_0, Emat)
                Vr[findex_0(i_s, strides, l), :, :] = Emat
            next_in_mrange(i_s, sizes)    # Increase i_s counter
        prev_V_L[:,:,::1] = curr_V_L
        prev_V_R[:,:,::1] = curr_V_R
    return E, Vl, Vr, CL_L, CL_R, CH_L, CH_R, array_strides_E, array_strides_L, array_strides_C, first_elements_E, first_elements_C



# -*- coding: utf-8 -*-
from unipolator.unitary_interpolation import *
from unipolator.symmetric_unitary_interpolation import *
from unipolator.hamiltonian_system import *
from unipolator.trotter_system import *
from unipolator.sym_trotter_system import *
#from .exp_and_log import *
#from .indexing import *
#from .caching import *
#from .blas_functions import *
#from .discrete_quantum import *



#cython: language_level=3
from libc.math cimport cos, sin, log
import numpy as np
#cimport numpy as np
from scipy.linalg.cython_blas cimport zgemm, zscal, zdscal, zaxpy, zcopy, zdotu
from scipy.linalg.cython_lapack cimport zheevd

cdef void AxB_elementwise_pointer(double complex *a0, double complex *b0, double complex *c0, int n2) nogil:
    cdef int i
    for i in range(n2):
        c0[i] = a0[i]*b0[i]

cdef void MM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *ori = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)
cdef void MM_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *ori = 'n'
    cdef double complex alpha_c
    cdef double complex beta = 0.0
    alpha_c.real = alpha
    alpha_c.imag = 0
    zgemm(ori, ori, &n, &n, &n, &alpha_c, b0, &n, a0, &n, &beta, c0, &n)
cpdef void MM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil: # C = A @ B
    # matrix multiply 2 square matrices A (n x n) and B (n x n) -> C = A @ B
    cdef char *ori = 'n'
    cdef double complex *a0=&A[0,0]
    cdef double complex *b0=&B[0,0]
    cdef double complex *c0=&C[0,0]
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    cdef int n
    n = A.shape[0]
    zgemm(ori, ori, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)

cdef void MMM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, double complex *e0, int n) nogil: # A @ B @ C = D
    cdef char *ori = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(ori, ori, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, e0, &n)   
    zgemm(ori, ori, &n, &n, &n, &alpha, c0, &n, e0, &n, &beta, d0, &n)
cdef void MMM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D, double complex[:,::1] E) nogil:  # A @ B @ C = D (helper variable E)
    cdef double complex *a0 = &A[0, 0]
    cdef double complex *b0 = &B[0, 0]
    cdef double complex *c0 = &C[0, 0]
    cdef double complex *d0 = &D[0, 0]
    cdef double complex *e0 = &E[0, 0]
    cdef int n = A.shape[0]
    MMM_cdot_pointer(a0, b0, c0, d0, e0, n)

cdef void DagM_M_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'n'
    cdef char *orientB = 'c'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(orientA, orientB, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)
cdef void DagM_M_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'n'
    cdef char *orientB = 'c'
    cdef double complex alpha_c
    cdef double complex beta = 0.0
    alpha_c.real = alpha
    alpha_c.imag = 0
    zgemm(orientA, orientB, &n, &n, &n, &alpha_c, b0, &n, a0, &n, &beta, c0, &n)
cpdef void DagM_M_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'n'
    cdef char *orientB = 'c'
    cdef double complex *a0=&A[0,0]
    cdef double complex *b0=&B[0,0]
    cdef double complex *c0=&C[0,0]
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    cdef int n = A.shape[0]
    zgemm(orientA, orientB, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)

cdef void M_DagM_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'c'
    cdef char *orientB = 'n'
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    zgemm(orientA, orientB, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)
cdef void M_DagM_cdot_scale_pointer(double alpha, double complex *a0, double complex *b0, double complex *c0, int n) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'c'
    cdef char *orientB = 'n'
    cdef double complex alpha_c
    cdef double complex beta = 0.0
    alpha_c.real = alpha
    alpha_c.imag = 0
    zgemm(orientA, orientB, &n, &n, &n, &alpha_c, b0, &n, a0, &n, &beta, c0, &n)
cpdef void M_DagM_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C) nogil:
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *orientA = 'c'
    cdef char *orientB = 'n'
    cdef double complex *a0=&A[0,0]
    cdef double complex *b0=&B[0,0]
    cdef double complex *c0=&C[0,0]
    cdef double complex alpha = 1.0
    cdef double complex beta = 0.0
    cdef int n = A.shape[0]
    zgemm(orientA, orientB, &n, &n, &n, &alpha, b0, &n, a0, &n, &beta, c0, &n)

# Triple products
cpdef void DagA_B_A_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D) nogil:
    # calculates the product: D = Dag(A) @ B @ A
    # uses C to store intermediate results
    DagM_M_cdot(A, B, C)
    MM_cdot(C,A, D)
cdef void DagA_B_A_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, int n) nogil:
    # calculates the product: D = Dag(A) @ B @ A
    # uses C to store intermediate results
    DagM_M_cdot_pointer(a0, b0, c0, n)
    MM_cdot_pointer(c0, a0, d0, n)

cpdef void A_B_DagA_cdot(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex[:,::1] D) nogil:
    # calculates the product: D = A @ B @ Dag(A)
    # uses C to store intermediate results
    MM_cdot(A, B, C)
    M_DagM_cdot(C, A, D)
cdef void A_B_DagA_cdot_pointer(double complex *a0, double complex *b0, double complex *c0, double complex *d0, int n) nogil:
    # calculates the product: D = A @ B @ Dag(A)
    # uses C to store intermediate results
    MM_cdot_pointer(a0, b0, c0, n)
    M_DagM_cdot_pointer(c0, a0, d0, n)

cdef double complex tr_dot_pointer(double complex *v0, double complex *w0, int n) nogil: # -i*E*amp*exp(-i*E*t)
    cdef double complex c
    cdef double abstr
    cdef int inc = 1
    cdef int i
    cdef double complex trAB = zdotu(&n, v0, &inc, w0, &n)
    for i in range(1,n):
        v0 += n
        w0 += 1
        trAB += zdotu(&n, v0, &inc, w0, &n)
    return trAB
cpdef double complex tr_dot(double complex[:,::1] V, double complex[:,::1] W, int n ):
    cdef double complex *v0 = &V[0,0]
    cdef double complex *w0 = &W[0,0]
    return tr_dot_pointer(v0, w0, n)

cdef double complex tr_dot_pointer_target_indexes(double complex *v0, double complex *w0, int n, int[::1] target_indexes) nogil: # -i*E*amp*exp(-i*E*t)
    cdef double complex c
    cdef int i, i_s
    cdef int inc = 1
    i = target_indexes[0]
    cdef double complex trAB = zdotu(&n, v0 + i * n, &inc, w0 + i, &n)
    for i_s in range(1,target_indexes.shape[0]):
        i = target_indexes[i_s]
        trAB += zdotu(&n, v0+i*n, &inc, w0+i, &n)
    return trAB
cpdef double complex tr_dot_target_indexes(double complex[:,::1] V, double complex[:,::1] W, int n, int[::1] target_indexes ):
    cdef double complex *v0 = &V[0,0]
    cdef double complex *w0 = &W[0,0]
    return tr_dot_pointer_target_indexes(v0, w0, n, target_indexes)

cdef double complex target_indexes_trace_pointer(double complex *v0, int n, int[::1] target_indexes) nogil:
    cdef int i
    cdef double complex c = 0
    for i in range(target_indexes.shape[0]):
        c += v0[i*n+i]
    return c

cpdef (int, int, int) c_eigh_lapack_workspace_sizes(double complex[:,::1] H):  # H will be returned as the
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef char *jobz = 'v' #eigenvectors and values -> v
    cdef char *uplo = 'l' # upper triangle
    cdef int n = H.shape[0]
    cdef double complex *h0=&H[0,0]
    cdef int info = 0
    cdef double[:] w = np.array([n], dtype=np.double)
    cdef double complex[:] work1 = np.empty([1], dtype=np.complex128)
    cdef double [:] rwork1 = np.empty([1], dtype=np.double)
    cdef int[:] iwork1 = np.empty([1], dtype=np.int32)
    cdef double complex *work0 = &work1[0]
    cdef double *rwork0 = &rwork1[0]
    cdef int *iwork0 = &iwork1[0]
    cdef int lrwork = -1
    cdef int lwork = -1
    cdef int liwork = -1
    zheevd(jobz, uplo, &n, h0, &n, &w[0], work0, &lwork, rwork0, &lrwork, iwork0, &liwork, &info)
    cdef int work = <int>(work1[0].real + 1e-9)
    cdef int rwork = <int>(rwork1[0] + 1e-9)
    cdef int iwork = <int>(iwork1[0] +1e-9)
    return work, rwork, iwork

cpdef void c_eigh_lapack(double complex[:,::1] H, double complex[:,::1] V, double[::1] E, int lwork, int lrwork, int liwork):  # H will be returned as the #int nwork, int nrwork, int niwork,
    # matrix multiply 2 square matrices A (n x n) and B (n x n)
    cdef int n = H.shape[0]
    V[:,:] = H
    cdef double complex *v0 = &V[0,0]
    cdef double *e0=&E[0]

    cdef char *jobz = 'v' #eigenvectors and values -> v
    cdef char *uplo = 'l' # upper triangle

    #cdef int lwork = 2*n + n*n+1
    cdef double complex[:] work = np.empty([lwork], dtype=np.complex128)
    cdef double complex *work0 = &work[0]
    #cdef int lrwork = 1 + 5*n + 2*n**2 +1
    cdef double[:] rwork = np.empty([lrwork], dtype=np.double)
    cdef double *rwork0 = &rwork[0]

    #cdef int liwork = 3 + 5*n +1
    cdef int[:] iwork = np.empty([liwork], dtype=np.int32)
    cdef int *iwork0 = &iwork[0]

    cdef int info = 0
    zheevd(jobz, uplo, &n, v0, &n, e0, work0, &lwork, rwork0, &lrwork, iwork0, &liwork, &info)
    #return info


###### Additions #######################################################################################################
cpdef void d_third_order_tensor_scale(double complex[:,:,::1] A, double d) nogil:
    cdef int s = A.shape[0]
    cdef int n = A.shape[1]
    cdef int snn = s*n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0,0]
    zdscal(&snn, &d, a0, &incz)

cpdef void d_mat_scale(double complex[:,::1] A, double d) nogil:
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    zdscal(&nn, &d, a0, &incz)

cpdef void c_mat_scale(double complex[:,::1] A, double complex c) nogil:
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    zscal(&nn, &c, a0, &incz)

cpdef void d_mat_add_first(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double d): # nogil:
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    C[:,:] = B
    d_mat_scale(C, d)
    cdef double complex *c0=&C[0,0]
    zcopy(&nn, c0, &incz, a0, &incz)

cpdef void c_mat_add_first(double complex[:,::1] A, double complex[:,::1] B, double complex[:,::1] C, double complex c): # nogil:
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    C[:,:] = B
    c_mat_scale(C, c)
    cdef double complex *c0=&C[0,0]
    zcopy(&nn, c0, &incz, a0, &incz)

cpdef void d_mat_add(double complex[:,::1] A, double complex[:,::1] B, double d) nogil: # A = A+B*d
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    cdef double complex *b0=&B[0,0]
    cdef double complex c = <double complex> d
    zaxpy(&nn, &c, b0, &incz, a0, &incz)

cdef void d_mat_add_pointer(double complex *a0, double complex *b0, double d, int n2) nogil: # A = A+B*d
    cdef int incz = 1
    cdef double complex c = <double complex> d
    zaxpy(&n2, &c, b0, &incz, a0, &incz)

cpdef void c_mat_add(double complex[:,::1] A, double complex[:,::1] B, double complex c) nogil: # A = A+B*c
    cdef int n = A.shape[0]
    cdef int nn = n*n
    cdef int incz = 1
    cdef double complex *a0=&A[0,0]
    cdef double complex *b0=&B[0,0]
    zaxpy(&nn, &c, b0, &incz, a0, &incz)




#cython: language_level=3
import numpy as np
cimport numpy as npc
from libc.math cimport fabs
from .exp_and_log cimport *
from .indexing cimport *
from .caching cimport *
from .blas_functions cimport *


# Unitary Interpolation
cdef class Sym_UI:
    # Initialize variables, to quickly calculate interpolations while minimizing memmory allocation overheads
    cdef double[::1] c_mins, c_maxs, dcs, das
    cdef long[::1] c_bins
    cdef int n_dims, d, d2, n_dims_1, n_dims_2, n_d_di_1, n_d_di, i
    cdef double[:, ::1] E_C, E
    cdef double complex[:,:,::1] Vr, Vl, CL_L, CH_L, CL_R, CH_R, dUl, dUr
    cdef double complex[:,::1] Ur, Ur1, Ur2, dUc
    cdef double complex[::1] expE
    cdef double *ei
    cdef double complex *expe
    cdef double complex *vl
    cdef double complex *vr
    cdef double complex *cl
    cdef double complex *cr
    cdef double complex *ur0
    cdef double complex *ur1
    cdef double complex *ur2
    cdef double complex *dul
    cdef double complex *dur
    cdef double complex *duc
    cdef long[::1] strides_L, strides_EC
    cdef long[:,::1] strides_E, strides_C
    cdef long[::1] location, d_location
    cdef double[::1] abs_alpha_rest, alpha
    cdef long[::1] first_elements_E, first_elements_C, L, 
    cdef long[::1] d_di
    def __cinit__(self, double complex[:,:,::1] H_s, double[::1] c_min_s, double[::1] c_max_s, long[::1] c_bins, long[::1] which_diffs = np.array([], dtype=long)):
        # Construct parameters
        self.n_dims = c_min_s.shape[0]
        self.n_dims_1 = self.n_dims - 1
        self.n_dims_2 = self.n_dims - 2
        self.d = H_s.shape[1]
        self.d2 = self.d * self.d
        if not H_s.shape[0] == self.n_dims + 1:
            print('Requires n+1 Hamiltonians for n dimensional interpolation. Check lenths of Hs, c_mins, c_maxs, c_bins')
            raise ValueError
        self.c_bins = np.empty(self.n_dims, dtype=long)
        for i in range(self.n_dims):
            if c_bins[i] > 0:
                self.c_bins[i] = c_bins[i]
            else:
                print('Need at least 1 bin per dimension, corrected bins['+str(i)+'] to 1.')
        self.c_mins, self.c_maxs, self.dcs, self.das = Bin_Parameters(c_min_s, c_max_s, self.c_bins)
        # Single Step Indexing Parameters  --> Add multiple indexes later
        self.location = np.empty(self.n_dims, dtype=long)
        self.abs_alpha_rest = np.empty(self.n_dims, dtype=np.double)
        self.d_location = np.empty(self.n_dims, dtype=long)
        self.alpha = np.empty(self.n_dims, dtype=np.double)
        if which_diffs.shape[0] == 0:
            self.d_di = np.arange(self.n_dims)
        else:
            self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

        # Construct interpolation grid points
        H_s2 = H_s.copy()
        d_third_order_tensor_scale(H_s2, 0.5 )
        U_grid2, cum_prod = Unitary_Grid(H_s2, self.c_mins, self.dcs, self.c_bins) # Half grid -> H_s2 = H_s / 2
        ## Construct interpolation cache
        self.E, self.Vl, self.Vr, self.CL_L, self.CL_R, self.CH_L, self.CH_R, self.strides_E, self.strides_L, self.strides_C, self.first_elements_E, self.first_elements_C = Create_Sym_Interpolation_Cache( U_grid2, cum_prod, self.c_bins)

        self.ei = &self.E_C[0, 0]
        self.vl = &self.Vl[0, 0, 0]
        self.vr = &self.Vr[0, 0, 0]
        self.cl = &self.CL_L[0, 0, 0]
        self.cr = &self.CL_R[0, 0, 0]
        self.Ur = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur0 = &self.Ur[0, 0]
        self.Ur1 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur1 = &self.Ur1[0, 0]
        self.Ur2 = np.empty([self.d, self.d], dtype=np.complex128)
        self.ur2 = &self.Ur2[0, 0]
        self.L = np.empty(self.n_dims, dtype=long)
        # Split into left and right side
        self.dUl = np.empty([self.n_dims, self.d, self.d], dtype=np.complex128)
        self.dul = &self.dUl[0, 0, 0]
        self.dUr = np.empty([self.n_dims, self.d, self.d], dtype=np.complex128)
        self.dur = &self.dUr[0, 0, 0]
        self.dUc = np.empty([self.d, self.d], dtype=np.complex128)
        self.duc = &self.dUc[0,0]
        self.expE = np.empty([self.d], dtype=np.complex128)
        self.expe = &self.expE[0]

    cdef single_parameters2oddgrid(self, double[::1] c):
        cdef long sum_location = 0
        cdef double alpha_max = 0.0
        cdef double alpha_rest
        cdef long i
        cdef long max_alpha_ind = 0
        cdef long max_vals
        for i in range(self.n_dims):
            # Transform
            self.alpha[i] = (c[i] - self.c_mins[i]) / self.dcs[i]
            # Round
            self.location[i] = <long> self.alpha[i]
            alpha_rest = self.alpha[i] - self.location[i]
            if alpha_rest > 0.5:
                self.location[i] += 1
                alpha_rest -= 1
            sum_location += self.location[i]

            self.abs_alpha_rest[i] = fabs(alpha_rest)
            self.d_location[i] = 1 if alpha_rest >= 0 else -1
            if self.alpha[i] < 0 or self.alpha[i] > self.c_bins[i]:
                print('Warning: These parameters lie outside of interpolation grid!')
                break
        if is_even(sum_location):
            max_alpha_ind = c_argmax(self.abs_alpha_rest, max_alpha_ind, alpha_max, self.n_dims)
            if self.location[max_alpha_ind] + self.d_location[max_alpha_ind] > self.c_bins[max_alpha_ind]:
                self.d_location[max_alpha_ind] = -1
            self.location[max_alpha_ind] += self.d_location[max_alpha_ind]
            self.d_location[max_alpha_ind] = - self.d_location[max_alpha_ind]
            self.abs_alpha_rest[max_alpha_ind] = 1-self.abs_alpha_rest[max_alpha_ind]
        for i in range(self.n_dims):
            self.d_location[i] -= 1
            self.d_location[i] >>= 1 # via bitshift operation
            #d_location[i] /= 2 #(d_location[i]-1)/2 ### in two steps with optimized in-place operations
        for i in range(self.n_dims):
            max_vals = self.location[i]+self.d_location[i]
            if max_vals > self.c_bins[i]-1:
                self.d_location[i] = -1

    def set_which_diffs(self, long[::1] which_diffs):
        self.d_di = which_diffs
        self.n_d_di_1 = self.d_di.shape[0] - 1
        self.n_d_di = self.d_di.shape[0]

    def get_single_param(self, double[::1] c): # To verify interpolation grid
        self.single_parameters2oddgrid(c)
        return np.asarray(self.location), np.asarray(self.d_location), np.asarray(self.abs_alpha_rest)

    def get_cache(self): # To verify interpolation grid
        return np.asarray(self.E_C), np.asarray(self.E), np.asarray(self.Vl), np.asarray(self.Vr), np.asarray(self.CL_L), np.asarray(self.CL_R), np.asarray(self.CH_L), np.asarray(self.CH_R), np.asarray(self.strides_EC), np.asarray(self.strides_E), np.asarray(self.strides_R), np.asarray(self.strides_C), np.asarray(self.first_elements_E), np.asarray(self.first_elements_C)

    cdef interpolate_single_u(self, double complex *u0):  #u0 => input the matrices for output
        cdef int i, j
        cdef int ind, indE
        if self.n_dims == 1:  # Shouldn't make any difference if you use symmetric 1D or non symmetric 1D interpolation
            ind = self.location[0] + self.d_location[0]
            self.vr = &self.Vr[ind, 0, 0]  # In 1D strides are 1
            self.vl = &self.Vl[ind, 0, 0]
            self.ei = &self.E[ind, 0]
            V_expE_W_pointer(self.vl, self.vr, self.ei, self.abs_alpha_rest[0], self.ur0, u0, self.d, self.d2)
        else:
            # From center outward
            self.L[0] = self.location[0] + self.d_location[0]
            for i in range(1, self.n_dims):
                self.L[i] = self.location[i]
            indE = findex_0(self.L, self.strides_E[0,:], self.n_dims)
            self.ei = &self.E[indE, 0]
            self.L[0] = self.location[0]
            self.L[1] += self.d_location[1]
            indE = findex_0(self.L, self.strides_E[1, :], self.n_dims) + self.first_elements_E[1]  # For E
            if self.d_location[0]:  ### Higher
                self.L[0] += -1
                ind = findex_0(self.L, self.strides_C[0, :], self.n_dims)
                self.cl = &self.CL_L[ind, 0, 0]
                self.cr = &self.CL_R[ind, 0, 0]
                self.L[0] += 1
            else:  ### Lower
                ind = findex_0(self.L, self.strides_C[0, :], self.n_dims)
                self.cl = &self.CH_L[ind, 0, 0]
                self.cr = &self.CH_R[ind, 0, 0]
            self.L[1] = self.location[1]
            V_expE_W_pointer(self.cl, self.cr, self.ei, self.abs_alpha_rest[0], self.ur0, self.ur1, self.d, self.d2)  # Save onto self.ur1
            # Now outward
            for k in range(self.n_dims_2):  # All except for the last element
                i = k + 1
                j = k + 2
                self.ei = &self.E[indE, 0]
                indE = findex_0(self.L, self.strides_E[j, :], self.n_dims) + self.first_elements_E[j]  # For E
                self.L[j] += self.d_location[j]
                if self.d_location[i]:  ### Higher
                    self.L[i] += -1
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.cl = &self.CL_L[ind, 0, 0]
                    self.cr = &self.CL_R[ind, 0, 0]
                    self.L[i] += 1 # Restore value
                else:  ### Lower
                    ind = findex_0(self.L, self.strides_C[i, :], self.n_dims) + self.first_elements_C[i]
                    self.cl = &self.CH_L[ind, 0, 0]
                    self.cr = &self.CH_R[ind, 0, 0]
                self.L[j] = self.location[j]
                #Continue here
                expE_V_expE_pointer(self.ei, self.ur1, self.abs_alpha_rest[i], self.expe, self.d)
                MMM_cdot_pointer(self.cl, self.ur1, self.cr, self.ur1, self.ur2, self.d) # Save back onto self.ur1
            # Last element
            self.ei = &self.E[indE, 0]
            self.L[self.n_dims_1] += self.d_location[self.n_dims_1]
            ind = findex_0(self.L, self.strides_L, self.n_dims)
            self.vl = &self.Vl[ind, 0, 0]
            self.vr = &self.Vr[ind, 0, 0]
            expE_V_expE_pointer(self.ei, self.ur1, self.abs_alpha_rest[self.n_dims_1], self.expe, self.d)
            MMM_cdot_pointer(self.vl, self.ur1, self.vr, u0, self.ur2, self.d)  # Save to u0

    cdef expmH_pointer(self, double[::1] c, double complex *u0):
        if not c.shape[0] == self.n_dims:
            raise ValueError('The coefficient c must be of size [interpolation_dimensions].')
        self.single_parameters2oddgrid(c)
        self.interpolate_single_u(u0)
    def expmH(self, double[::1] c, double complex[:,::1] U):
        cdef double complex *u0 = &U[0, 0]
        self.expmH_pointer(c, u0)



#!/usr/bin/python3
import os
from setuptools import setup, find_packages, Extension
import numpy as np

name = 'unipolator'
annotate = False

try:
    from Cython.Build import cythonize
    print('Compiling via cython')
except ImportError:
    print('Not using Cythonize, due to an ImportError')
    cythonize = None

# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

# make a list of the .pyx files in the os.join.path("src",name) directory
#pyx_files = [os.path.splitext(fn)[0] for fn in os.listdir(os.path.join('src', name)) if fn.endswith(".pyx")] 
pyx_files = ['blas_functions', 'exp_and_log', 'indexing', 'caching', 'unitary_interpolation', 'symmetric_unitary_interpolation', 'hamiltonian_system', 'trotter_system', 'sym_trotter_system']

extra_compile_args = ["-O3", "-march=native"]
include_dirs = [np.get_include()]
extensions = [Extension(name+'.'+filename, [ os.path.join('src', name, filename+'.pyx')], include_dirs=include_dirs, extra_compile_args=extra_compile_args) for filename in pyx_files]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

CYTHONIZE = True if cythonize is not None else False
if CYTHONIZE:
    compiler_directives = {'initializedcheck': False, 'boundscheck': False, 'wraparound': False, 'language_level': 3, "embedsignature": True, "cdivision": True, "nonecheck" : False}
    extensions = cythonize(extensions, language_level = "3", annotate=annotate, compiler_directives=compiler_directives ) #gdb_debug=True, 
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-dev.txt") as fp:
    dev_requires = fp.read().strip().split("\n")

setup(   
    name = "unipolator",
    #zip_safe = False,
    version         = "0.2.6",
    author = "Michael Schilling",
    author_email = "michael@ntropic.de",
    description  = "Unipolator allows for n dimensional unitary interpolation, and the calculation of propagators using unitary interpolation. Speeds up your propagators for linear quantum systems.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/Ntropic/unipolator",
    download_url = "https://github.com/Ntropic/unipolator/archive/refs/tags/v0.2.6.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    ext_modules = extensions,
    install_requires = install_requires,
    extras_require = {
        "dev": dev_requires,
        "docs": ["sphinx", "sphinx-rtd-theme"]
    },
)



cpdef Bin_Parameters(double[::1] c_mins, double[::1] c_maxs, long[::1] c_bins)

cpdef void next_in_mrange(long[:] A, long[:] n_s) nogil

cpdef reverse_cum_prod(long[:] s, int n)

cpdef int findex_0(long[:] i, long[:] cum_prod, int n)

cpdef int findex(long[:] i, long[:] cum_prod, int n, int a) nogil

cpdef flat_index_sizes_E(long[::1] c_bins)

cpdef flat_index_sizes_V(long[::1] c_bins)

cpdef int is_odd(int num) nogil

cpdef void is_odd_array(long[::1] num, int n) nogil

cpdef int is_even(int num) nogil

cpdef void is_even_array(long[::1] num, int n) nogil

cpdef void asym_sign(double[::1] nums, long[::1] signs, int n) nogil

cpdef void asym_sign_2D(double[:,::1] nums, long[:,::1] signs, int n0, int n1) nogil

cpdef void rounder(double a, int b) nogil

cpdef void rounder_array(double[::1] a, long[::1] b, int n) nogil

cpdef void rounder_array_2d(double[:,::1] a, long[:,::1] b, int n0, int n1) nogil

cpdef int is_in_interval(double[::1] c, long[::1] maxs, int n) nogil

cpdef int int_sum(long[::1] i_s, int n) nogil

cpdef int int_max(long[::1] i_s, int n) nogil

cpdef int int_prod(long[::1] i_s, int n) nogil

cpdef int_prod_array(long[:,::1] i_s, int n)
cpdef int_prod_asym_array(long[:,::1] i_s)

cpdef void c_abs_sum(double[::1] c, double[::1] abs_c, int n) nogil

cpdef void c_int_sum(long[::1] c, int sum_c, int n) nogil

cpdef void dvec_minus_ivec(double[::1] c, double[::1] dvec, long[:] ivec, int n) nogil

cpdef void elementwise_grid_parameters(double[::1] c, double[::1] c_min, double[::1] dc, long[::1] c_bins, double[::1] alpha, int sum_location, long[::1] location, double[::1] alpha_rest, double[::1] abs_alpha_rest, long[::1] d_location) nogil

cpdef int c_argmax(double[::1] c, int max_ind, double c_max, int n) nogil

#cpdef Parameters2OddGrid(double[::1] alpha, double[::1] c, double[::1] c_min, double[::1] dc, int[::1] c_bins)

